#  Copyright (C) 1999-2008
#  Smithsonian Astrophysical Observatory, Cambridge, MA, USA
#  For conditions of distribution and use, see copyright notice in "copyright"

package provide DS9 1.0

proc CATCDS {varname} {
    upvar #0 $varname var
    global $varname
    global cat

    global debug
    if {$debug(tcl,cat)} {
	puts "CATCDS $varname"
    }

    global http

    set var(reader) CATCDSReader

    switch -- $cat(server) {
	cds {set site "vizier.u-strasbg.fr"}
	sao {set site "vizier.cfa.harvard.edu"}
	cadc {set site "vizier.hia.nrc.ca"}
	adac {set site "vizier.nao.ac.jp"}
	iucaa {set site "vizier.iucaa.ernet.in"}
	inasan {set site "vizier.inasan.ru"}
	bejing {set site "data.bao.ac.cn"}
	cambridge {set site "apm20.ast.cam.ac.uk"}
	jac {set site "www.ukirt.jach.hawaii.edu"}
    }

    set cgidir {viz-bin}
    set script {asu-tsv}

    set var(url) "http://$site/$cgidir/$script"
    
    # defaults
    set query {}
    append query "-source=$var(catalog)&"

    # source
    append query "-c.ra=$var(x)&"
    append query "-c.dec=$var(y)&"
    switch -- $var(sky) {
	fk4 {append query "-c.eq=B1950&"}
	fk5 {append query "-c.eq=J2000&"}
	icrs {append query "-c.eq=J2000&"}
	galactic {append query "-c.eq=Gal&"}
    }

    switch -- $var(rformat) {
	degrees {append query "-c.bd=$var(width)x$var(height)&"}
	arcmin {append query "-c.bm=$var(width)x$var(height)&"}
	arcsec {append query "-c.bs=$var(width)x$var(height)&"}
    }

    # output
    if {$var(allrows)} {
	append query "-out.max=100000&"
    } else {
	append query "-out.max=$var(max)&"
    }
    if {$var(allcols)} {
	append query "-out.all&"
    }

    set out [http::formatQuery "Tab-Separated-Values"]
    append query "-out.form=$out&"

    set cols [http::formatQuery _RAJ,_DEJ]
    append query "-out.add=$cols&"
    append query "-oc.form=dec"

    set var(query) $query

    CATLoad $varname
}

proc CATCDSReader {t sock token} {
    upvar #0 $t T
    global $t

    global debug
    if {$debug(tcl,cat)} {
#	puts "CATCDSReader"
    }

    set result 0

    if { ![info exists ${t}(state)]  } {
	set T(state) 0
    }

    switch -- $T(state) {
	0 {
	    # init db
	    fconfigure $sock -blocking 1
	    set T(Nrows) 0
	    set T(Ncols) 0
	    set T(Header) {}
	    set T(HLines) 0

	    set T(state) 1
	}

	1 {
	    # process header
	    incr ${t}(HLines)
	    set n $T(HLines)
	    if {[gets $sock line] == -1} {
		set T(state) -1
		set T(HLines) [expr $T(HLines) - 1]
		set T(Nrows) 0
		set T(Ncols) 0
		return
	    }

	    set result [string length "$line"]
	    set T(H_$n) $line
	    if {[regexp -- {^ *(-)+ *(\t *(-)+ *)*} $line]} {
		# remove units line
		unset T(H_$n)
		incr ${t}(HLines) -1
		incr n -1
		set T(H_$n) $line
		
		# clean up header column name
		set hh $T(H_[expr $n-1])
		regsub -all {\[} $hh {} hh
		regsub -all {\]} $hh {} hh
		set T(H_[expr $n-1]) $hh

		# cols
		set T(Header) [split $T(H_[expr $n-1]) "\t"]
		set T(Dashes) [split $T(H_$n) "\t"]
		set T(Ndshs) [llength $T(Dashes)]
		
		starbase_colmap $t
		set T(state) 2
	    }
	}

	2 { 
	    # process table
	    if {[gets $sock line] == -1} {
		set T(state) 0
	    } else {
		set result [string length "$line"]
		set line [string trim $line]

		if {$line != {}} {
		    # check for beginning of another table
		    if {[string range $line 0 0] == "#"} {
			set T(state) 3
			return $result
		    }

		    # check for garbage at start of line
		    if {![string is double [lindex $line 0]]} {
			set T(state) 3
			return $result
		    }

		    # ok, save it
		    incr ${t}(Nrows)
		    set r $T(Nrows)

		    set NCols [starbase_ncols $t]
		    set c 1
		    foreach val [split $line "\t"] {
			set T($r,$c) $val
			incr c
		    }
		    for {} {$c <= $NCols} {incr c} {
			set T($r,$c) {}
		    }
		}
	    }
	}

	3 {
	    # finished, eat everything else
	    if {[gets $sock line] == -1} {
		set T(state) 0
	    }
	}
    }

    return $result
}

proc CATCDSAck {varname} {
    upvar #0 $varname var
    global $varname

    set msg {Acknowledgments for CDS

This research has made use of the VizieR catalogue access tool, CDS,
Strasbourg, France. VizieR is a joint effort of CDS (Centre de Données
astronomiques de Strasbourg) and ESA-ESRIN (Information Systems
Division).
}
    SimpleTextDialog ${varname}ack [msgcat::mc {Acknowledgment}] 80 10 insert top $msg
}
