#  Copyright (C) 1999-2008
#  Smithsonian Astrophysical Observatory, Cambridge, MA, USA
#  For conditions of distribution and use, see copyright notice in "copyright"

package provide DS9 1.0

proc HVCancel {varname} {
    upvar #0 $varname var
    global $varname

    global debug
    if {$debug(tcl,hv)} {
	puts "HVCancel"
    }

    # set state to 0 so that we don't process the finish proc
    set var(active) 0

    # stop any refresh
    if {$var(refresh,id)>0} {
	after cancel $var(refresh,id)
	set var(refresh,id) 0
    }

    # analysis
    if {$var(analysis)} {
	AnalysisTaskEnd $var(analysis,which) $var(analysis,i)
	HVSetAnalysis $varname 0 {} 0
    }

    # clean up
    HVClearTmpFile $varname

    if [info exists var(token)] {
	http::reset $var(token)
    }

    if [info exists var(widget)] {
	$var(widget) configure -cursor {}
    }
}

proc HVDestroy {varname} {
    upvar #0 $varname var
    global $varname

    global debug
    if {$debug(tcl,hv)} {
	puts "HVDestroy"
    }

    HVCancel $varname

    # clear the widge and all images
    $var(widget) clear

    # clear image cache
    foreach x [array names $varname "images,*"] {
	image delete $var($x)
	unset ${varname}($x)
    }

    # clear cache
    HVClearCache $varname

    # destroy the window and menubar
    if {[winfo exists $var(top)]} {
	destroy $var(top)
	destroy $var(mb)
    }

    # clear varname
    unset $varname
}

proc HVReset {varname} {
    upvar #0 $varname var
    global $varname

    set var(active) 0

    if {[info exists var(token)]} {
	global debug
	if {$debug(tcl,hv)} {
	    puts "HVReset ***cleanup***"
	}
	http::cleanup $var(token)
	unset var(token)
    }
}

proc HVDone {varname} {
    upvar #0 $varname var
    global $varname

    HVStatus $varname {}
    HVReset $varname
}

proc HVCancelled {varname} {
    upvar #0 $varname var
    global $varname

    HVStatus $varname {}
    HVReset $varname
}

proc HVError {varname err} {
    upvar #0 $varname var
    global $varname

    global debug
    if {$debug(tcl,hv)} {
	puts "HVError $err"
    }

    HVReset $varname
    Error $err
}

proc HVStatus {varname message} {
    upvar #0 $varname var
    global $varname

    set var(status) $message
}

# this is the main entry point, everybody calls here

proc HVLoadURL {varname url query} {
    upvar #0 $varname var
    global $varname

    # this assumes the url has been already resolved

    global debug
    if {$debug(tcl,hv)} {
	puts "HVLoadURL $varname $url $query"
    }

    # do we have anything?
    if {$url == {}} {
	return
    }

    HVStatus $varname {}

    # set the cursor to busy
    if [info exists var(widget)] {
	$var(widget) configure -cursor watch
    }

    # parse url
    ParseURL $url r
    if {$debug(tcl,hv)} {
	puts "HVLoadURL |$r(scheme)|$r(authority)|$r(path)|$r(query)|$r(fragment)|$query|"
    }

    switch -- $r(scheme) {
	file -
	{} {HVProcessURLFile $varname $url $query r}
	ftp {HVProcessURLFTP $varname $url $query r}
	http {HVProcessURLHTTP $varname $url $query r}
	default {HVError $varname "[msgcat::mc {Sorry, DS9 does not support}] $r(scheme)"}
    }
}

proc HVProcessURLFile {varname url query rr} {
    upvar #0 $varname var
    global $varname

    upvar $rr r
    
    global ds9

    global debug
    if {$debug(tcl,hv)} {
	puts "HVProcessURLFile"
    }

    if [file exists $r(path)] {
	if [file isdirectory $r(path)] {
	    HVSetURL $varname $url {} {}
	    HVSetResult $varname 200 "text/html"
	    HVSetData $varname \
		[HVFileHtmlList $r(path) [HVDirList $r(path)]] {}

	    set var(delete) 0
	    HVParse $varname
	} else {
	    HVSetURL $varname $url {} $r(fragment)
	    set var(delete) 0
	    HVLoadFile $varname $r(path)
	}
    }
}

proc HVProcessURLFTP {varname url query rr} {
    upvar #0 $varname var
    global $varname

    upvar $rr r
    
    global ds9

    global debug
    if {$debug(tcl,hv)} {
	puts "HVProcessURLFTP"
    }

    set fn "$ds9(tmpdir)/[file tail $r(path)]"
    set ftp [ftp::Open $r(authority) "ftp" "-ds9@" -mode passive]
    if {$ftp > -1} {
	# first try to get as file
	set ftp::VERBOSE $debug(tcl,ftp)
	set "ftp::ftp${ftp}(Output)" FTPLog
	ftp::Type $ftp binary
	if [ftp::Get $ftp $r(path) "$fn"] {
	    ftp::Close $ftp

	    if {$debug(tcl,hv)} {
		puts "HVProcessURLFTP get $fn"
	    }
	    HVSetURL $varname $url {} $r(fragment)

	    set var(delete) 1
	    HVLoadFile $varname "$fn"

	    HVClearTmpFile $varname
	} else {
	    # from the prev attempt
	    catch {file delete -force "$fn"}

	    if {$debug(tcl,hv)} {
		puts "HVProcessURLFTP list"
	    }

	    # now as a directory
	    set list [ftp::List $ftp $r(path)]
	    ftp::Close $ftp

	    HVSetURL $varname $url {} {}
	    HVSetResult $varname 200 "text/html"
	    HVSetData $varname [HVFTPHtmlList $r(authority) $r(path) $list] {}

	    set var(delete) 0
	    HVParse $varname
	}
    }
}

proc HVProcessURLHTTP {varname url query rr} {
    upvar #0 $varname var
    global $varname

    upvar $rr r
    
    global ds9
    global http

    global debug
    if {$debug(tcl,hv)} {
	puts "HVProcessURLHTTP"
    }

    # stop any refresh
    if [info exists ${varname}(refresh,id)] {
	if {$var(refresh,id)>0} {
	    after cancel $var(refresh,id)
	}
    }

    # do we already have it in cache?
    if {[info exists ${varname}(cache,file,$url,$query)]} {

	# has it expired?
	if {($var(cache,expire,$url,$query) == 0) ||
	    ($var(cache,expire,$url,$query) > [clock seconds])} {

	    # just in case
	    if [file exists $var(cache,file,$url,$query)] {
		# ok, to it
		if {$debug(tcl,hv)} {
		    puts "HVProcessURLHTTP found $url at $var(cache,file,$url,$query)"
		}

		HVSetURL $varname $url $query $r(fragment)
		set var(delete) 0
		HVSetResult $varname 200 $var(cache,mime,$url,$query)
		HVSetData $varname {} $var(cache,file,$url,$query)

		HVParse $varname

		return
	    }
	}

	# expired or invalid, clean up
	if {$debug(tcl,hv)} {
	    puts "HVProcessURLHTTP expired or invalid $var(cache,file,$url,$query)"
	}
	catch {file delete $var(cache,file,$url,$query)}
	unset var(cache,file,$url,$query)
	unset var(cache,mime,$url,$query)
	unset var(cache,expire,$url,$query)
    }

    HVSetURL $varname $url $query $r(fragment)
    HVSetResult $varname {} {}
    HVSetData $varname {} {}

    set var(ch) {}

    # do we have html? if so, use a var
    ParseURL $url r

    # geturl as file
    set var(fn) [tmpnam ds9 .http]
    if [catch {open "$var(fn)" w} ${varname}(ch)] {
	HVError $varname "[msgcat::mc {Unable to open file}] $var(fn)"
	return
    }

    # disable timeouts for analysis
    set timeout $http(timeout)
    if {$var(analysis)} {
	set timeout 0
    }

    # alwasy async
    if {![catch {set var(token) [http::geturl $url \
				     -timeout $timeout \
				     -progress [list HVProgress $varname] \
				     -binary 1 \
				     -headers "[ProxyHTTP]" \
				     -query "$query" \
				     -channel $var(ch) \
				     -command \
				     [list HVProcessURLHTTPFinish $varname]]
    }]} {
	set var(active) 1
	set var(delete) 1
    } else {
	catch {close $var(ch)}
	HVError $varname "[msgcat::mc {Unable to locate URL}] $url"
    }
}

proc HVProcessURLHTTPFinish {varname token} {
    upvar #0 $varname var
    global $varname

    global debug
    if {$debug(tcl,hv)} {
	puts "HVProcessURLHTTPFinish"
    }

    catch {close $var(ch)}

    if {!($var(active))} {
	HVCancelled $varname
	return
    }

    upvar #0 $token t

    # Code
    set var(code) [http::ncode $token]

    # Meta
    set var(meta) $t(meta)

    # Cache defaults
    set var(cache) 1
    set var(cache,images) 1
    set var(expire) 0

    HVParseMeta $varname

    # Log it
    HTTPLog $token

    # Result?
    switch -- $var(code) {
	200 -
	203 -
	404 -
	503 {
	    if {$var(cache)} {
		if {$debug(tcl,hv)} {
		    puts "HVProcessURLHTTPFinish cacheing:$var(url),$var(query):$var(fn)"
		}
		set url $var(url)
		set query $var(query)
		set var(cache,file,$url,$query) $var(fn)
		set var(cache,mime,$url,$query) $var(mime)
		set var(cache,expire,$url,$query) $var(expire)
		set var(delete) 0
	    }
	    HVParse $varname
	    HVDone $varname
	}

	201 -
	300 -
	301 -
	302 -
	303 -
	305 -
	307 {
	    foreach {name value} $var(meta) {
		if {[regexp -nocase ^location$ $name]} {
		    global debug
		    if {$debug(tcl,hv)} {
			puts "HVProcessURLHTTPFinish redirect $var(code) to $value"
		    }
		    # clean up and resubmit
		    http::cleanup $token
		    unset var(token)

		    HVClearTmpFile $varname

		    if [info exists var(widget)] {
			HVLoadURL $varname [$var(widget) resolve $value] {}
		    } else {
			HVLoadURL $varname $value {}
		    }
		}
	    }
	}

	default {HVError $varname "HTTP [msgcat::mc {Error}] $var(code)"}
    }
}

proc HVParseMeta {varname} {
    upvar #0 $varname var
    global $varname

    global debug
    if {$debug(tcl,hv)} {
	puts "HVParseMeta: $var(meta)"
    }

    foreach {name value} $var(meta) {
	switch -- [string tolower $name] {
	    content-type {
		regexp -nocase {([^;]*);?(.*)} $value foo \
		    ${varname}(mime) ${varname}(mime,param)
		set var(mime) [string tolower $var(mime)]
	    }
	    content-length {}
	    content-encoding {
		switch -- [string tolower $value] {
		    gzip -
		    x-gzip {set var(encoding) gzip}
		    bzip2 {set var(encoding) bzip2}
		    compress -
		    Z {set var(encoding) compress}
		    pack -
		    z {set var(encoding) pack}
		    default {}
		}
	    }
	    content-transfer-encoding {
		switch -- [string tolower $value] {
		    binary -
		    base64 {set var(transfer) [string tolower $value]}
		    default {}
		}
	    }

	    refresh {
		set f [split $value \;]
		set var(refresh,time) [lindex $f 0]
		set var(refresh,url) [string range [lindex $f 1] 4 end]
		if {$var(refresh,url) != {} & $var(refresh,time) != {}} {
		    set var(refresh,id) [after [expr $var(refresh,time)*1000] "HVLoadURL $varname \{$var(refresh,url)\} {}"]
		} else {
		    set var(refresh,id) 0
		}
	    }
	    expires {
		if {[catch {set ss [clock scan $value]}]} {
		    set var(cache) 0
		} else {
		    set var(cache) 1
		    set var(expire) $ss
		}
	    }
	    cache-control {
		foreach cc [split $value {,}] {
		    foreach {nn vv} [split $cc {=}] {
			switch $nn {
			    public {set var(cache) 1}
			    private {set var(cache) 1}
			    no-cache {set var(cache) 0}
			    no-store {set var(cache) 1}

			    s-maxage -
			    min-fresh -
			    max-age {
				set var(cache) 1
				set var(expire) \
				    [expr [file mtime $var(fn)]+$vv]
			    }
			    max-stale {}
			    no-transform {}
			    only-if-cached {}
			    cache-extension {}
			    
			    must-revalidate {}
			    proxy-revalidate {}
			}
		    }
		}
	    }
	    pragma {
		switch $value {
		    no-cache {set var(cache) 0}
		}
	    }
	    last-modified {
	    }
	    if-none-match {
	    }
	    set-cookie {
	    }
	}
    }

    if {$debug(tcl,hv)} {
	puts "HVParseMeta Content-Type:$var(mime):$var(mime,param):"
	puts "HVParseMeta Content-Encoding:$var(encoding):"
	puts "HVParseMeta Content-Transfer-Encoding:$var(transfer):"
	puts "HVParseMeta Refresh:$var(refresh,time):$var(refresh,url):"
	puts "HVParseMeta Cache:$var(cache)"
    }
}

proc HVLoadFile {varname path} {
    upvar #0 $varname var
    global $varname

    global debug
    if {$debug(tcl,hv)} {
	puts "HVLoadFile $path"
    }

    HVSetResult $varname 200 {}
    HVSetData $varname {} $path

    # content-encoding
    switch -- [string tolower [file extension $path]] {
	.gz {
	    set path [file rootname $path]
	    set var(encoding) gzip
	}
	.bz2 {
	    set path [file rootname $path]
	    set var(encoding) bzip2
	}
	.Z {
	    set path [file rootname $path]
	    set var(encoding) compress
	}
	.z {
	    set path [file rootname $path]
	    set var(encoding) pack
	}
    }

    switch -- [string tolower [file extension $path]] {
	.html -
	.htm {set var(mime) "text/html"}
	.gif {set var(mime) "image/gif"}
	.jpeg -
	.jpg {set var(mime) "image/jpeg"}
	.tiff -
	.tif {set var(mime) "image/tiff"}
	.png {set var(mime) "image/png"}
	.bmp {set var(mime) "image/bmp"}
	.ppm {set var(mime) "image/ppm"}
	.xbm {set var(mime) "image/xbm"}

	.fits -
	.fit -
	.fts {set var(mime) "image/fits"}

	.ftz -
	.fits.gz -
	.fgz {
	    set var(mime) "image/fits"
	    set var(encoding) "gzip"
	}

	.text -
	.txt {set var(mime) "text/plain"}
	.multi {
	    set var(mime) "multipart/mixed"
	    set var(xpa,target) "*:*"
	    if [file exists "$path"] {
		set ch [open "$path" r]
		if {[gets $ch line] >= 0} {
		    set var(mime,param) "Content-Type: multipart/mixed; Boundary=[string range $line 2 end]"
		}
		catch {close $ch}
	    }
	}
	default {
	    switch -- $var(encoding) {
		gzip -
		bzip2 -
		compress -
		pack {set var(mime) "application/octet-stream"}
		default {set var(mime) "text/plain"}
	    }
	}
    }

    HVParse $varname
}

proc HVParse {varname} {
    upvar #0 $varname var
    global $varname

    global debug
    if {$debug(tcl,hv)} {
	puts "HVParse"
    }

    switch -- $var(mime) {
	"multipart/alternative" -
	"multipart/parallel" -
	"multipart/digest" -
	"multipart/related" -
	"multipart/signed" -
	"multipart/encrypted" -
	"multipart/report" {}

	"multipart/x-mixed-replace" -
	"multipart/mixed" {HVParseMulti $varname}

	default {HVParseSingle $varname}
    }

    if {$var(analysis)} {
	AnalysisTaskEnd $var(analysis,which) $var(analysis,i)
	HVSetAnalysis $varname 0 {} 0
    }
}

proc HVParseMulti {varname} {
    upvar #0 $varname var
    global $varname

    global debug
    if {$debug(tcl,hv)} {
	puts "HVParseMulti"
    }

    # do it now, to be restored later
    if [info exists var(index)] {
	incr ${varname}(index)
	set var(index,$var(index)) "$var(url) $var(query)"
	set index $var(index)
    }

    set fn $var(fn)
    set del $var(delete)

    if [file exists "$var(fn)"] {
	if [catch {open "$var(fn)" r} ch] {
	    HVError $varname "[msgcat::mc {Unable to open file}] $var(fn)"
	    return
	}
    }

    set boundary [HVParseMimeParam $varname "boundary"]
    if {![string compare "$boundary" {}]} {
	HVError $varname [msgcat::mc {Invalid formated multipart/mixed mime type message}]
	return
    }

    set state 1
    set var(ch) {}

    HVSetResult $varname 200 {}
    HVSetData $varname {} {}

    while {[gets $ch line] >= 0} {
	if {$debug(tcl,hv)} {
#	    puts "HVParseMulti $state:$line"
	}

	switch -- $state {
	    1 {
		# boundary
		if {![string compare "--$boundary" $line]} {
		    set state 2
		}
	    }
	    2 {
		# header
		if {[string length $line] == 0} {
		    HVParseMeta $varname

		    # save to a file
		    set var(fn) [tmpnam ds9 .http]
		    set var(delete) 1
		    if [catch {open "$var(fn)" w} ${varname}(ch)] {
			HVError $varname "[msgcat::mc {Unable to open file}] $var(fn)"
			return
		    }
		    switch $var(transfer) {
			binary -
			base64 {
			    fconfigure $var(ch) \
				-translation binary -encoding binary
			}
		    }

		    set state 3
		} else {
		    if {[regexp -nocase {^([^:]+):(.+)$} $line x key value]} {
			lappend var(meta) $key [string trim $value]
		    }
		}
	    }
	    3 {
		# body
		if {![string compare "--$boundary" $line]} {
		    catch {close $var(ch)}
		    HVParseSingle $varname
		    HVClearTmpFile $varname

		    set var(ch) {}

		    HVSetResult $varname 200 {}

		    # we want to preserve var(text)
		    # HVSetData $varname {} {}
		    set var(data) {}
		    set var(fn) {}

		    set state 2

		} elseif {![string compare "--$boundary--" $line]} {
		    catch {close $var(ch)}
		    catch {close $ch}

		    HVParseSingle $varname
		    HVClearTmpFile $varname

		    # reset file values
		    set var(fn) $fn
		    set var(delete) $del

		    if [info exists var(index)] {
			# reset index
			set var(index) $index
			HVClearIndex $varname $index
		    }

		    return

		} else {
		    switch $var(transfer) {
			binary {puts -nonewline $var(ch) $line}
			base64 {
			    puts -nonewline $var(ch) [base64::decode $line]
			}
			default {puts $var(ch) $line}
		    }
		}
	    }
	}
    }

    # clean up
    catch {close $ch}
    set var(fn) $fn
}

proc HVParseSingle {varname} {
    upvar #0 $varname var
    global $varname

    global debug
    if {$debug(tcl,hv)} {
	puts "HVParseSingle $var(url)"
    }

    switch -- $var(mime) {
	"text/html" {HVParseHTML $varname}
	"text/plain" {HVParseText $varname}

	"image/gif" -
	"image/jpeg" -
	"image/tiff" -
	"image/png" -
	"image/bmp" -
	"image/ppm" -
	"image/x-portable-pixmap" -
	"image/xbm" -
	"image/x-xbitmap" {HVParseImg $varname}

	"image/fits" -
	"application/fits" {HVParseFITS $varname}

	"application/fits-image" -
	"application/fits-table" -
	"application/fits-group" {HVParseFITS $varname}

	"image/x-fits" -
	"binary/x-fits" -
	"application/x-fits" {HVParseFITS $varname}

	"image/x-gfits" -
	"binary/x-gfits" -
	"image/gz-fits" -
	"application/x-gzip" -
	"display/gz-fits" {
	    set var(encoding) gzip
	    HVParseFITS $varname
	}

	"image/bz2-fits" -
	"display/bz2-fits" {
	    set var(encoding) bzip2
	    HVParseFITS $varname
	}

	"image/x-cfits" -
	"binary/x-cfits" {
	    set var(encoding) compress
	    HVParseFITS $varname
	}

	"image/x-zfits" -
	"binary/x-zfits" {
	    set var(encoding) pack
	    HVParseFITS $varname
	}

	"image/fits-hcompress" -
	"image/x-fits-h" {HVParseSave $varname}

	"application/octet-stream" {
	    # one last chance to grap it as a fits file

	    ParseURL $var(url) r
	    set path [file tail $r(path)]  

	    # set content-encoding
	    switch -- [file extension $path] {
		.gz {
		    set path [file rootname $path]
		    set var(encoding) gzip
		}
		.bz2 {
		    set path [file rootname $path]
		    set var(encoding) bzip2
		}
		.Z {
		    set path [file rootname $path]
		    set var(encoding) compress
		}
		.z {
		    set path [file rootname $path]
		    set var(encoding) pack
		}
		default {set var(encoding) {}}
	    }

	    # set Content-Type
	    switch -- [file extension $path] {
		.fits -
		.fit -
		.fts {
		    set var(mime) "image/fits"
		    set var(mime,param) {}
		    HVParseFITS $varname
		}

		.ftz -
		.fgz {
		    set var(mime) "image/fits"
		    set var(mime,param) {}
		    set var(encoding) "gzip"
		    HVParseFITS $varname
		}
		default {HVParseSave $varname}
	    }
	}

	"x-xpa/xpaget" {}
	"x-xpa/xpaset" {HVParseXPASet $varname}
	"x-xpa/xpainfo" {}
	"x-xpa/xpaaccess" {}

	default {HVParseSave $varname}
    }
}

proc HVParseText {varname} {
    upvar #0 $varname var
    global $varname

    global debug
    if {$debug(tcl,hv)} {
	puts "HVParseText"
    }

    if {[string length $var(data)] == 0} {
	if [file exists "$var(fn)"] {
	    if [catch {open "$var(fn)" r} ch] {
		HVError $varname "[msgcat::mc {Unable to open file}] $var(fn)"
		return
	    }
	    set var(data) [read $ch]
	    close $ch
	}
    }

    if {$var(analysis)} {
	if {$debug(tcl,hv)} {
	    puts "HVParseText Analysis"
	}
	AnalysisProcessGetURL $var(analysis,which) $var(analysis,i) $var(data)
    } else {
	append var(text) $var(data)
	set var(data) \
	    "<html>\n<body>\n<pre>\n$var(text)\n</pre>\n</body>\n</html>"
	HVSetResult $varname 200 "text/html"
	HVParseHTML $varname
    }
}

proc HVParseHTML {varname} {
    upvar #0 $varname var
    global $varname

    global debug
    if {$debug(tcl,hv)} {
	puts "HVParseHTML"
    }

    if {[string length $var(data)] == 0} {
	if [file exists "$var(fn)"] {
	    if [catch {open "$var(fn)" r} ch] {
		HVError $varname "[msgcat::mc {Unable to open file}] $var(fn)"
		return
	    }
	    set var(data) [read $ch]
	    close $ch
	}
    }

    # check for HTTP-EQUIV
    if {[HVParseHTTP-EQUIV $varname]} {
	return
    }

    # figure out the base
    # we don't want any query or fragments
    ParseURL $var(url) r

    set base {}
    # scheme
    if {[string length $r(scheme)] != 0} {
	append base "$r(scheme)://"
    }
    # authority
    if {[string length $r(authority)] != 0} {
	append base "$r(authority)"
    }
    # path
    if {[string length $r(path)] != 0} {
	append base "$r(path)"
    } else {
	append base "/"
    }
    # query
    if {[string length $r(query)] != 0} {
	append base "?$r(query)"
    }
    $var(widget) config -base $base

    if {$debug(tcl,hv)} {
	DumpURL r
	puts "HVParseHTML base [$var(widget) cget -base]"
    }

    # we have a valid html
    $var(widget) clear

    # fix forms with no action
    HVFixHTMLForm $varname

    # and now, parse it
    $var(widget) parse $var(data)

    HVGotoHTML $varname
}

proc HVParseHTTP-EQUIV {varname} {
    upvar #0 $varname var
    global $varname

    global debug
    if {$debug(tcl,hv)} {
	puts "HVParseHTTP-EQUIV"
    }

    if {[regexp {<HEAD>(.*)</HEAD>} $var(data) foo head]} {
	if {[regexp {<META HTTP-EQUIV} $head foo]} {
	    
	    # do it now, to be restored later
	    if [info exists var(index)] {
		incr ${varname}(index)
		set var(index,$var(index)) "$var(url) $var(query)"
		set index $var(index)
	    }

	    set fn $var(fn)
	    set del $var(delete)
	    set data $var(data)

	    # init
	    HVSetResult $varname 200 {}
	    HVSetData $varname {} {}

	    # header
	    foreach {line} [split $head "\n"] {
		set line [string trim $line]
		set line [string trim $line "<>"]
		set exp {META HTTP-EQUIV=(.*) CONTENT=(.*)}
		if [regexp $exp $line foo key value] {
		    if {$key != {} && $value != {}} {
			lappend var(meta) \
			    [string trim $key {"}] [string trim $value {"}]
		    }
		}
	    }	    

	    # data
	    if {[regexp {<BODY>(.*)</BODY>} $data foo body]} {
		set var(fn) [tmpnam ds9 .http]
		set var(delete) 1
		if [catch {open "$var(fn)" w} ch] {
		    HVError $varname "[msgcat::mc {Unable to open file}] $var(fn)"
		    return
		}

		HVParseMeta $varname

		switch $var(transfer) {
		    binary {
			fconfigure $ch -translation binary -encoding binary
			puts -nonewline $ch $body
		    }
		    base64 {
			fconfigure $ch -translation binary -encoding binary
			puts -nonewline $ch [base64::decode $body]
		    }
		    default {puts -nonewline $ch $body}
		}
		catch {close $ch}

		HVParse $varname
		HVClearTmpFile $varname
	    }

	    # reset file values
	    set var(fn) $fn
	    set var(delete) $del
	    
	    if [info exists var(index)] {
		# reset index
		set var(index) $index
		HVClearIndex $varname $index
	    }

	    return 1
	}
    }

    return 0
}

proc HVParseImg {varname} {
    upvar #0 $varname var
    global $varname

    global debug
    if {$debug(tcl,hv)} {
	puts "HVParseImg $var(url)"
    }

    # fake html 
    set html \
	"<html>\n<body>\n<img src=\"$var(fn)\" border=0>\n</body>\n</html>"

    HVSetURL $varname "$var(url)" {} {}
    HVSetResult $varname 200 "text/html"
    HVSetData $varname $html "$var(fn)"

    set var(cache,images) 0

    $var(widget) clear
    $var(widget) config -base {}
    $var(widget) parse $var(data)

    HVGotoHTML $varname
}

proc HVParseFITS {varname} {
    upvar #0 $varname var
    global $varname

    global ds9
    global debug

    if {$var(delete) && $var(save)} {
	switch -- $var(encoding) {
	    gzip {FileLast savefitsfbox "ds9.fits.gz"}
	    bzip2 {FileLast savefitsfbox "ds9.fits.bz2"}
	    compress {FileLast savefitsfbox "ds9.fits.Z"}
	    pack {FileLast savefitsfbox "ds9.fits.z"}
	    default {FileLast savefitsfbox "ds9.fits"}
	}

	set fn [SaveFileDialog savefitsfbox]
	if {[string length "$fn"] != 0} {
	    if {![catch {file rename -force "$var(fn)" "$fn"}]} {
		set var(fn) "$fn"
		set var(delete) 0
	    }
	}
    }

    switch -- $var(frame) {
	new {
	    set ds9(display,user) tile
	    DisplayMode
	    CreateFrame
	}
	current {}
    }

    StartLoad
    global loadParam
    set loadParam(load,type) allocgz
    set loadParam(load,layer) {}
    set loadParam(file,type) fits
    set loadParam(file,mode) "$var(file,mode)"
    set loadParam(file,name) "$var(fn)"
    set loadParam(file,fn) $loadParam(file,name)

    # may have to convert the file, based on content-encoding
    switch -- "$var(encoding)" {
	bzip2 {
	    catch {set ch [open "| bunzip2 < $var(fn) " r]}
	    set loadParam(load,type) channel
	    set loadParam(channel,name) $ch
	}
	compress {
	    catch {set ch [open "| uncompress < $var(fn) " r]}
	    set loadParam(load,type) channel
	    set loadParam(channel,name) $ch
	}
	pack {
	    catch {set ch [open "| pcat $var(fn) " r]}
	    set loadParam(load,type) channel
	    set loadParam(channel,name) $ch
	}
    }

    ProcessLoad
    FinishLoad

    HVClearTmpFile $varname
    HVClearAll $varname
    HVUpdateDialog $varname
}

proc HVParseSave {varname} {
    upvar #0 $varname var
    global $varname

    global debug
    if {$debug(tcl,hv)} {
	puts "HVParseSave"
    }

    set fn [HVParseMimeParam $varname "name"]
    if {$fn == {}} {
	ParseURL $var(url) r
	set fn [file tail $r(path)]
    }
    FileLast savefitsfbox $fn
    set fn [SaveFileDialog savefitsfbox]
    if {[string length "$fn"] != 0} {
	if {![catch {file rename -force "$var(fn)" "$fn"}]} {
	    set var(delete) 0
	}
    }

    HVClearAll $varname
    HVUpdateDialog $varname
}

proc HVParseXPASet {varname} {
    upvar #0 $varname var
    global $varname
    global ds9

    global debug
    if {$debug(tcl,hv)} {
	puts "HVParseXPASet: [HVParseMimeParam $varname paramlist]"
    }

    if [info exists var(xpa,target)] {
	set target $var(xpa,target)
    } else {
	set target [HVParseMimeParam $varname target]
    }

    if {$target == "$ds9(title)" ||
	$target == "DS9:*" ||
	$target == "DS9:$ds9(title)" ||
	$target == "*:$ds9(title)" ||
	$target == "*:*" ||
	$target == [XPAMethod]} {

	HVXPASet $var(fn) [HVParseMimeParam $varname paramlist]
    } else {
	HVError $varname "[msgcat::mc {Unable to match target with XPA Mime request}] $url"
    }
}

proc HVGotoHTML {varname} {
    upvar #0 $varname var
    global $varname

    incr ${varname}(index)
    set var(index,$var(index)) "$var(url) $var(query)"

    global debug
    if {$debug(tcl,hv)} {
	puts "HVGotoHTML $var(index) |$var(url)|$var(query)|$var(fragment)|"
    }

    if {[string length $var(fragment)] != 0} {
	if {$debug(tcl,idletasks)} {
	    puts "HVGotoHTML"
	}
	update idletasks

	$var(widget) yview $var(fragment)
    } else {
	$var(widget) yview moveto 0
    }

    HVUpdateDialog $varname
}

proc HVClearCache {varname} {
    upvar #0 $varname var
    global $varname

    global debug
    if {$debug(tcl,hv)} {
	puts "HVClearCache $varname"
    }

    foreach x [array names $varname "cache,file,*"] {
	catch {file delete $var($x)}
    }
    foreach x [array names $varname "cache,*"] {
	unset ${varname}($x)
    }
}

proc HVClearIndex {varname n} {
    upvar #0 $varname var
    global $varname

    global debug
    if {$debug(tcl,hv)} {
	puts "HVClearIndex $varname $n"
    }

    foreach x [array names $varname "index,*"] {
	set f [split $x ,]
	if {[lindex $f 1] > $n} {
	    unset ${varname}($x)
	}
    }
    set var(index) $n
}

proc HVClearTmpFile {varname} {
    upvar #0 $varname var
    global $varname

    global debug
    if {$debug(tcl,hv)} {
	puts "HVClearTmpFile"
    }

    if {$var(delete) && [string length "$var(fn)"] != 0} {
	if [file exists "$var(fn)"] {
	    if {$debug(tcl,hv)} {
		puts "HVClearTmpFile delete $var(fn)"
	    }
	    file delete "$var(fn)"
	}
	set var(fn) {}
	set var(delete) 0
    }
}

proc HVUpdateDialog {varname} {
    upvar #0 $varname var
    global $varname

    # in case we've set the cursor
    $var(widget) configure -cursor {}

    global debug
    if {$debug(tcl,hv)} {
	puts "HVUpdateDialog\n"
    }

    set id $var(index)
    set id [incr id -1]
    if {[info exists ${varname}(index,$id)]} {
	$var(mb).view entryconfig [msgcat::mc {Back}] -state normal    
    } else {
	$var(mb).view entryconfig [msgcat::mc {Back}] -state disabled
    }

    set id $var(index)
    set id [incr id 1]
    if {[info exists ${varname}(index,$id)]} {
	$var(mb).view entryconfig [msgcat::mc {Forward}] -state normal    
    } else {
	$var(mb).view entryconfig [msgcat::mc {Forward}] -state disabled
    }
}

proc HVRefresh {varname} {
    upvar #0 $varname var
    global $varname

    global debug
    if {$debug(tcl,hv)} {
	puts "HVRefresh"
    }

    set var(delete) 0
    HVParse $varname
}

proc HVProgress {varname token totalsize currentsize} {
    upvar #0 $varname var
    global $varname

    if {$totalsize != 0} {
	HVStatus $varname "$currentsize bytes of $totalsize bytes [expr int(double($currentsize)/$totalsize*100)]%"
    } else {
	HVStatus $varname "$currentsize bytes"
    }
}

proc HVFTPHtmlList {host path list} {
    global debug
    if {$debug(tcl,hv)} {
	puts "HVFTPHtmlList $host $path"
    }
    if {[string range $path end end] != "/"} {
	append path {/}
    }

    set html {}
    append html "<html>\n"
    append html "<head>\n"
    append html "<title>Index of $path on $host</title>\n"
    append html "</head>\n"
    append html "<body>\n"
    append html "<h1>Index on $path on $host</h1>\n"
    append html "<hr>\n"
    append html "<pre>\n"
    foreach l $list {
	switch -- [llength $l] {
	    8 {set offset 4}
	    9 {set offset 5}
	    10 {set offset 4}
	    11 {set offset 5}
	    default {set offset 5}
	}

	set ii [lindex $l [expr $offset+3]]
	switch -- [string range $l 0 0] {
	    d {
		set new "<a href=\"ftp://$host$path$ii/\">$ii</A>"
	    }
	    l {
		set new "<a href=\"ftp://$host$path$ii\">$ii</A>"
	    }
	    default {
		set new "<a href=\"ftp://$host$path$ii\">$ii</A>"
	    }
	}

	regsub $ii $l $new l
	append html "$l\n"
    }
    append html "</pre>\n"
    append html "</hr>\n"
    append html "</body>\n"

    return $html
}

proc HVFileHtmlList {path list} {
    global debug
    if {$debug(tcl,hv)} {
	puts "HVFileHtmlList $path"
    }

    if {[string range $path end end] != "/"} {
	append path {/}
    }

    set html {}
    append html "<html>\n"
    append html "<head>\n"
    append html "<title>Index of $path</title>\n"
    append html "</head>\n"
    append html "<body>\n"
    append html "<h1>Index on $path</h1>\n"
    append html "<hr>\n"
    append html "<pre>\n"
    foreach l $list {
	switch -- [llength $l] {
	    8 {set offset 4}
	    9 {set offset 5}
	    10 {set offset 4}
	    11 {set offset 5}
	    default {set offset 5}
	}

	set ii [lindex $l [expr $offset+3]]
	switch -- [string range $l 0 0] {
	    d {
		set new "<a href=\"file:$path$ii/\">$ii</A>"
	    }
	    l {
		set new "<a href=\"file:$path$ii\">$ii</A>"
	    }
	    default {
		set new "<a href=\"file:$path$ii\">$ii</A>"
	    }
	}

	regsub $ii $l $new l
	append html "$l\n"
    }
    append html "</pre>\n"
    append html "</hr>\n"
    append html "</body>\n"

    return $html
}

proc HVDirList {path} {
    global debug
    if {$debug(tcl,hv)} {
	puts "HVDirList $path"
    }
    return [split [exec ls -l $path] \n]
}

proc HVSetURL {varname url query fragment} {
    upvar #0 $varname var
    global $varname

    global debug
    if {$debug(tcl,hv)} {
	puts "HVSetURL $url $query $fragment"
    }

    set var(url) $url
    set var(query) $query
    set var(fragment) $fragment
}

proc HVSetResult {varname code mime} {
    upvar #0 $varname var
    global $varname

    global debug
    if {$debug(tcl,hv)} {
	puts "HVSetResult $code $mime"
    }

    set var(code) $code
    set var(meta) {}
    set var(mime) $mime
    set var(mime,param) {}
    set var(cache) 0
    set var(cache,images) 1
    set var(expire) 0
    set var(encoding) {}
    set var(transfer) {}
    set var(refresh,time) 0
    set var(refresh,url) {}
    set var(refresh,id) 0
}

proc HVSetData {varname data fn} {
    upvar #0 $varname var
    global $varname

    global debug
    if {$debug(tcl,hv)} {
	puts "HVSetData $fn"
    }

    set var(data) $data
    set var(text) {}
    set var(fn) "$fn"
}

proc HVSetAnalysis {varname aa which ii} {
    upvar #0 $varname var
    global $varname

    global debug
    if {$debug(tcl,hv)} {
	puts "HVSetAnalysis"
    }

    set var(analysis) $aa
    set var(analysis,which) $which
    set var(analysis,i) $ii
}

proc HVClearAll {varname} {
    upvar #0 $varname var
    global $varname

    global debug
    if {$debug(tcl,hv)} {
	puts "HVClearAll"
    }

    HVSetURL $varname {} {} {}
    HVSetResult $varname {} {}
    HVSetData $varname {} {}
    HVSetAnalysis $varname 0 {} 0
}

# CallBacks

proc HVImageCB {varname args} {
    upvar #0 $varname var
    global $varname

    global ds9

    global debug
    if {$debug(tcl,hv)} {
	puts "HVImageCB args: $varname $args"
    }

    set url [lindex $args 0]
    if {$debug(tcl,hv)} {
	puts "HVImageCB $url"
    }

    # do we have anything?
    if {[string length $url] == 0} {
	return
    }

    # do we have a width/height?
    set aa [lindex $args 3]
    set width [HVattrs width $aa 0]
    set height [HVattrs height $aa 0]
    set src [HVattrs src $aa 0]

    # check for percent (100%) in width/height
    if {![string is integer $width]} {
	set width 0
    }
    if {![string is integer $height]} {
	set height 0
    }

    if {$debug(tcl,hv)} {
	puts "HVImageCB $width $height $src"
    }

    # we have a problem in that htmlwidget will not properly resolve a windows
    # file name, there for we may have a bad file name url
    # double check with the src attribute
    global tcl_platform
    switch $tcl_platform(platform) {
	unix {}
	windows {
	    ParseURL $url r

	    switch -- $r(scheme) {
		{} -
		file {
		    if {![file exists $url]} {
			# bad, try src
			if [file exists $src] {
			    set url $src
			}
		    }
		}
	    }
	}
    }

    set img [HVImageURL $varname $url $width $height]

    if {[string length $img] != 0} {
	return $img
    } else {
	if {$debug(tcl,hv)} {
	    puts "HVImageCB FAIL $url"
	}
	return $var(images,gray)
    }
}

proc HVImageURL {varname url width height} {
    upvar #0 $varname var
    global $varname

    global http

    global debug
    if {$debug(tcl,hv)} {
	puts "HVImageURL $varname $url $width $height"
    }

    # do we already have the image?
    if {[info exists ${varname}(images,$url)]} {
	if {$debug(tcl,hv)} {
	    puts "HVImageURL found image a $url"
	}
	return $var(images,$url)
    }

    ParseURL $url r

    set fn {}
    switch -- $r(scheme) {
	{} -
	file {
	    if [file exists $r(path)] {
		# can't use -file for zvfs
		# catch {image create photo -file $r(path)} img
		set ch [open $r(path) r]
		fconfigure $ch -translation binary -encoding binary
		set dd [read $ch]
		close $ch
		unset ch

		catch {image create photo -data "$dd"} img
		unset dd
	    }
	}
	ftp {
	    set fn [tmpnam ds9 [file extension $r(path)]]
	    set ftp [ftp::Open $r(authority) "ftp" "-ds9@" -mode passive]
	    if {$ftp > -1} {
		set ftp::VERBOSE $debug(tcl,ftp)
		set "ftp::ftp${ftp}(Output)" FTPLog
		ftp::Type $ftp binary
		if [ftp::Get $ftp $r(path) "$fn"] {
		    ftp::Close $ftp

		    if {[file size "$fn"] == 0} {
			catch {file delete -force "$fn"}
			return {}
		    }
		    if {[catch {image create photo -file "$fn"} img]} {
			catch {file delete -force "$fn"}
			return {}
		    }
		}
	    }
	}
	http {
	    set ch {}

	    set fn [tmpnam ds9 [file extension $r(path)]]

	    for {set ii 5} {$ii>0} {incr ii -1} {
		if [catch {open "$fn" w} ch] {
		    HVError $varname "[msgcat::mc {Unable to open file}] $fn"
		    return {}
		}

		if [catch {set token \
			       [http::geturl $url \
				    -timeout $http(timeout) \
				    -progress [list HVProgress $varname] \
				    -channel $ch \
				    -binary 1 \
				    -headers "[ProxyHTTP]" \
				   ]}] {
		    # if there is a problem, just bail
		    set ii 0
		    continue
		}

		catch {close $ch}

		upvar #0 $token t
		set code [http::ncode $token]
		set meta $t(meta)

		# result?
		switch -- $code {
		    200 -
		    203 -
		    503 {set ii 0}
		
		    201 -
		    300 -
		    301 -
		    302 -
		    303 -
		    305 -
		    307 {
			foreach {name value} $meta {
			    if {[regexp -nocase ^location$ $name]} {
				global debug
				if {$debug(tcl,hv)} {
				    puts "HVImageURL redirect $code to $value"
				}
				# clean up and resubmit
				set url $value
				http::cleanup $token
				catch {file delete -force "$fn"}
			    }
			}
		    }

		    default {
			# in general, we don't want to know about this
			if {$debug(tcl,hv)} {
			    puts "HVImageURL HTTP Error: $code"
			}
			set ii 0
		    }
		}
	    }

	    http::cleanup $token

	    if {[file size "$fn"] == 0} {
		catch {file delete -force "$fn"}
		return {}
	    }
	    if {[catch {image create photo -file "$fn"} img]} {
		catch {file delete -force "$fn"}
		return {}
	    }
	}
    }

    # do we have an img?
    if {![info exists img]} {
	return {}
    }
    if {$debug(tcl,hv)} {
	puts "HVImageURL got image $img"
    }

    # adjust image size if needed
    if {$width != 0 || $height != 0} {
	set iw [image width $img]
	set ih [image height $img]

	set doit 1
	# check for one dimension of 0. calculate to maintain aspect ratio
	if {$width == 0} {
	    set width [expr $iw*$height/$ih]

	    # see if we have a bad resample dimension
	    set wf [expr double($iw)*$height/$ih]
	    if {$width != $wf} {
		if {$debug(tcl,hv)} {
		    puts "HVImageURL abort resample image $img width $wf"
		}
		set doit 0
	    }
	}
	if {$height == 0} {
	    set height [expr $ih*$width/$iw]

	    # see if we have a bad resample dimension
	    set hf [expr double($ih)*$width/$iw]
	    if {$height != $hf} {
		if {$debug(tcl,hv)} {
		    puts "HVImageURL abort resample image $img height $hf"
		}
		set doit 0
	    }
	}

	# check to resample
	if {$doit && ($width != $iw || $height != $ih)} {
	    if {$debug(tcl,hv)} {
		puts "HVImageURL resample image $iw->$width $ih->$height"
	    }

	    set img2 \
		[image create photo -width $width -height $height]
	    if {[catch {blt::winop image resample $img $img2 box} ]} {
		# just use existing img
		if {$debug(tcl,hv)} {
		    puts "HVImageURL error resample image $img"
		}
	    } else {
		set tmp $img
		set img $img2
		catch {image delete $tmp}
	    }
	}
    }

    # delete any tmp files
    if {"$fn" != {}} {
	catch {file delete -force "$fn"}
    }

    if {$var(cache,images)} {
	set var(images,$url) $img
    }

    return $img
}

proc HVFontCB {varname sz args} {
    upvar #0 $varname var
    global $varname

    global debug
    if {$debug(tcl,hv)} {
	puts "HVFont $varname $sz $args"
    }

    set family helvetica
    set param {}
    set size 12

    foreach f [concat [lindex $args 0]] {
	switch -- $f {
	    fixed {set family courier}
	    default {append param "$f "} 
	}
    }
    
    switch -- $sz {
	1 {set size 6}
	2 {set size 8}
	3 {set size 10}
	4 {set size 12}
	5 {set size 16}
	6 {set size 20}
	7 {set size 24}
	default {set size 12}
    }

#	1 {set size 8}
#	2 {set size 10}
#	3 {set size 12}
#	4 {set size 16}
#	5 {set size 20}
#	6 {set size 24}
#	7 {set size 36}
#	default {set size 16}

    incr size $var(font,size)

    return "$family $size $param"
}

proc HVNoScriptCB {varname n tag args} {
    upvar #0 $varname var
    global $varname

    global debug
    if {$debug(tcl,hv)} {
	puts "HVNoScript $varname $n $tag $args"
    }
}

proc HVScriptCB {varname args} {
    upvar #0 $varname var

    global debug
    if {$debug(tcl,hv)} {
	puts "HVScriptCB $varname $args"
    }
}

proc HVFrameCB {varname args} {
    upvar #0 $varname var
    global $varname

    if {$debug(tcl,hv)} {
	puts "HVFrameCB $varname $args"
    }
}

proc HVAppletCB {varname w args} {
    upvar #0 $varname var
    global $varname

    global debug
    if {$debug(tcl,hv)} {
	puts "HVAppletCB $varname $w $args"
    }
}

proc HVParseMimeParam {varname key} {
    upvar #0 $varname var
    global $varname

    foreach {pp} [split $var(mime,param) {;}] {
	set id [string first {=} $pp]
	set name [string trim [string range $pp 0 [expr $id-1]]]
	set value [string trim [string range $pp [expr $id+1] end]]
	if {![string compare -nocase $name $key]} {
	    return [string trim $value {"'}]
	}
    }

    return {}
}
