#  Copyright (C) 1999-2008
#  Smithsonian Astrophysical Observatory, Cambridge, MA, USA
#  For conditions of distribution and use, see copyright notice in "copyright"

package provide DS9 1.0

# canvas

proc CreateCanvas {} {
    global ds9
    global canvas

    LayoutViewAdjust diff
    set ww [expr $canvas(width)+$diff(x)]
    set hh [expr $canvas(height)+$diff(y)]

    global debug
    if {$debug(tcl,layout)} {
	puts "CreateCanvas $canvas(width) $canvas(height) ${ww}x${hh}"
    }

    set ds9(image) [frame $ds9(main).f -relief ridge -bd 1]
    set ds9(canvas) [canvas $ds9(image).c -width $ww -height $hh \
			 -highlightthickness 0 -insertofftime 0]

    # needed to realize window so Layout routines will work
    pack $ds9(canvas) -fill both -expand true
    grid $ds9(image)

    # we need to manually set the plot bg color, i.e. use the default bg
    set canvas(bg) [$ds9(main) cget -background]
}

proc InitCanvas {} {
    global ds9

    # Bindings
    bind $ds9(canvas) <Tab> [list NextFrame]
    bind $ds9(canvas) <Shift-Tab> [list PrevFrame]
    switch $ds9(wm) {
	x11 {bind $ds9(canvas) <ISO_Left_Tab> [list PrevFrame]} 
	aqua {}
	win32 {}
    }

    bind $ds9(canvas) <Configure> [list ConfigureView]
    switch $ds9(wm) {
	x11 -
	aqua {
	    bind $ds9(canvas) <Enter> [list focus $ds9(canvas)]
	    bind $ds9(canvas) <Leave> [list focus {}]
	}
	win32 {
	    bind $ds9(canvas) <KeyPress> [list focus $ds9(canvas)]
	    bind $ds9(canvas) <Button> [list focus $ds9(canvas)]
	    bind $ds9(canvas) <Leave> [list focus {}]
	}
    }

    # backward compatible bindings
    switch $ds9(wm) {
	x11 -
	win32 {
	    bind $ds9(canvas) <Button-3> {Button3Canvas %x %y}
	    bind $ds9(canvas) <B3-Motion> {Motion3Canvas %x %y}
	    bind $ds9(canvas) <ButtonRelease-3> {Release3Canvas %x %y}
	} 
	aqua {
	    # swap button-2 and button-3 on the mighty mouse
	    bind $ds9(canvas) <Button-2> {Button3Canvas %x %y}
	    bind $ds9(canvas) <B2-Motion> {Motion3Canvas %x %y}
	    bind $ds9(canvas) <ButtonRelease-2> {Release3Canvas %x %y}

	    # x11 command key emulation
	    bind $ds9(canvas) <Command-Button-1> {Button3Canvas %x %y}
	    bind $ds9(canvas) <Command-B1-Motion> {Motion3Canvas %x %y}
	    bind $ds9(canvas) <Command-ButtonRelease-1> {Release3Canvas %x %y}
	}
    }

    # freeze
    bind $ds9(canvas) <f> {ToggleBindEvents}
}

proc Button3Canvas {x y} {
    global ds9

    global debug
    if {$debug(tcl,events)} {
	puts "Button3Canvas"
    }

    set ds9(b3) 1
    ButtonColormap $x $y
}

proc Motion3Canvas {x y} {
    global debug
    if {$debug(tcl,events)} {
	puts "Motion3Canvas"
    }

    MotionColormap $x $y
}

proc Release3Canvas {x y} {
    global ds9
    global debug
    if {$debug(tcl,events)} {
	puts "Release3Canvas"
    }

    set ds9(b3) 0
    ReleaseColormap $x $y
}

proc UnBindEventsCanvas {} {
    global ds9

    foreach f $ds9(active) {
	UnBindEventsFrame $f
    }
}

proc BindEventsCanvas {} {
    global ds9
    global current

    switch -- $ds9(display,mode) {
	single -
	blink {BindEventsFrame $current(frame)}
	tile {
	    foreach f $ds9(active) {
		BindEventsFrame $f
	    }
	}
    }
}

# view

proc ConfigureView {} {
    global canvas
    global debug
    if {$debug(tcl,layout)} {
	puts "ConfigureView $canvas(width) $canvas(height)"
    }

    LayoutView
}

proc UpdateView {} {
    global debug
    global canvas
    if {$debug(tcl,layout)} {
	puts "UpdateView $canvas(width) $canvas(height)"
    }

    SetWatchCursor
    LayoutDS9
    UnsetWatchCursor
}

proc LayoutDS9 {} {
    global ds9
    global view
    global canvas
    global colorbar
    global graph

    global debug
    if {$debug(tcl,layout)} {
	puts "LayoutDS9 $canvas(width) $canvas(height)"
    }

    # save current size
    set wo [winfo width $ds9(top)]
    set ho [winfo height $ds9(top)]

    # adjust window size
    LayoutViewAdjust diff
    set ww [expr $canvas(width)+$diff(x)]
    set hh [expr $canvas(height)+$diff(y)]

    # determine how much to change
    set wc [winfo width $ds9(canvas)]
    set hc [winfo height $ds9(canvas)]

    set wt [winfo width $ds9(top)]
    set ht [winfo height $ds9(top)]
    
    set w [expr $ww - $wc + $wt]
    set h [expr $hh - $hc + $ht]

    # change window size
    wm geometry $ds9(top) ${w}x${h}

    # special case
    # if the window size is the same, we still need to run LayoutView
    # however, if there is a change, configure will run it for us
    if {$wo == $w && $ho == $h} {
	LayoutView
    }
}

proc LayoutView {} {
    global ds9
    global view
    global canvas
    global graph
    global colorbar

    global debug
    if {$debug(tcl,layout)} {
	puts "LayoutView old $canvas(width) $canvas(height)"
    }

    # calc sizes
    LayoutViewAdjust diff
    set canvas(width) [expr [winfo width $ds9(canvas)]-$diff(x)]
    set canvas(height) [expr [winfo height $ds9(canvas)]-$diff(y)]
    if {$debug(tcl,layout)} {
	puts "LayoutView new $canvas(width) $canvas(height)"
    }

    # layout
    switch $view(layout) {
	vertical {LayoutViewVert}
	horizontal -
	default {LayoutViewHorz}
    }

    LayoutInfoPanel
    LayoutFrames
    LayoutColorbar
    LayoutGraphs
}

proc LayoutViewAdjust {varname} {
    upvar $varname var

    global view
    global colorbar
    global graph
    global current

    set var(x) 0
    set var(y) 0
    if {$view(colorbar)} {
	set h [expr $colorbar(gap)+$colorbar(height)]
	if {$h>$var(y)} {
	    set var(y) $h
	}
    }
    if {$view(graph,vert)} {
	set w [expr $graph(gap,x)+$graph(size)]
	if {$w>$var(x)} {
	    set var(x) $w
	}

	set h $graph(gap,y)
	if {$h>$var(y)} {
	    set var(y) $h
	}
    }
    if {$view(graph,horz)} {
	set w $graph(gap,x)
	if {$w>$var(x)} {
	    set var(x) $w
	}
	
	if {$view(colorbar)} {
	    set h [expr $colorbar(gap)+$colorbar(height)]
	} else {
	    set h $graph(gap,y)
	}
	incr h $graph(size)
	if {$h>$var(y)} {
	    set var(y) $h
	}
    }

    global debug
    if {$debug(tcl,layout)} {
	puts "LayoutViewAdjust $var(x) $var(y)"
    }
}

proc LayoutViewReset {} {
    global ds9
    global canvas

    global debug
    if {$debug(tcl,layout)} {
	puts "LayoutViewReset"
    }

    grid rowconfigure $ds9(main) 2 -weight 0
    grid columnconfigure $ds9(main) 0 -weight 0

    grid rowconfigure $ds9(main) 0 -weight 0
    grid columnconfigure $ds9(main) 1 -weight 0

    grid forget $ds9(panel)
    grid forget $ds9(buttons)
    grid forget $ds9(image)
    pack forget $ds9(info)
    pack forget $ds9(panner)
    pack forget $ds9(magnifier)
}

proc LayoutViewHorz {} {
    global ds9
    global current
    global view

    # canvas
    grid rowconfigure $ds9(main) 2 -weight 1
    grid columnconfigure $ds9(main) 0 -weight 1
    grid $ds9(image) -row 2 -column 0 -sticky news

    # buttons
    if {$view(buttons)} {
	grid $ds9(buttons) -row 1 -sticky ew -columnspan 3
    } else {
	grid forget $ds9(buttons)
    }

    # info panel
    if {$view(info) || $view(magnifier) || $view(panner)} {
	grid $ds9(panel) -row 0 -column 0 -sticky ew -columnspan 3
    } else {
	grid forget $ds9(panel)
    }

    if {$view(info)} {
	pack $ds9(info) -side left -padx $ds9(internalpad) \
	    -pady $ds9(internalpad) -fill x -expand true
    } else {
	pack forget $ds9(info)
    }

    if {$view(panner)} {
	pack $ds9(panner) -side right -padx $ds9(internalpad) \
	    -pady $ds9(internalpad)
    } else {
	pack forget $ds9(panner)
    }

    if {$view(magnifier)} {
	pack $ds9(magnifier) -side right -padx $ds9(internalpad) \
	    -pady $ds9(internalpad)
	if {$view(panner)} {
	    pack $ds9(magnifier) -before $ds9(panner)
	}
    } else {
	pack forget $ds9(magnifier)
    }
}

proc LayoutViewVert {} {
    global ds9
    global current
    global view

    # canvas
    grid rowconfigure $ds9(main) 0 -weight 1
    grid columnconfigure $ds9(main) 1 -weight 1
    grid $ds9(image) -row 0 -column 1 -sticky news

    # info panel
    if {$view(info) || $view(magnifier) || $view(panner)} {
	grid $ds9(panel) -row 0 -column 0 -sticky ns
    } else {
	grid forget $ds9(panel)
    }

    if {$view(magnifier)} {
	pack $ds9(magnifier) -side top -padx $ds9(internalpad) \
	    -pady $ds9(internalpad)
    } else {
	pack forget $ds9(magnifier)
    }

    if {$view(info)} {
	pack $ds9(info) -side top -padx $ds9(internalpad) \
	    -pady $ds9(internalpad) -fill y -expand true
	if {$view(magnifier)} {
	    pack $ds9(info) -after $ds9(magnifier)
	}
    } else {
	pack forget $ds9(info)
    }

    if {$view(panner)} {
	pack $ds9(panner) -side bottom -padx $ds9(internalpad) \
	    -pady $ds9(internalpad)
    } else {
	pack forget $ds9(panner)
    }

    # buttons
    grid forget $ds9(buttons)
}

proc LayoutColorbar {} {
    global ds9
    global view
    global canvas
    global colorbar
    global current

    set xx 0
    set yy [expr $canvas(height) + $colorbar(gap)]

    set colorbar(width) $canvas(width)

    switch -- $ds9(visual) {
	pseudocolor {
	    colorbar configure -x $xx -y $yy -width $colorbar(width) -anchor nw
	}
	truecolor {
	    colorbar configure -x $xx -y $yy -width $colorbar(width) -anchor nw
	    colorbarrgb configure -x $xx -y $yy -width $colorbar(width)\
		-anchor nw
	}
    }
}

proc LayoutGraphs {} {
    global ds9
    global view
    global canvas
    global colorbar
    global graph

    if {$view(graph,horz)} {
	set xx 0
	if {$view(colorbar)} {
	    set yy [expr $canvas(height)+$colorbar(gap)+$colorbar(height)]
	} else {
	    set yy [expr $canvas(height)+$graph(gap,y)]
	}
	set ww [expr $canvas(width)+$graph(gap,x)]

	if {$graph(horz,id) == 0} {
	    set graph(horz,id) [$ds9(canvas) create window $xx $yy \
				    -window $ds9(graph,horz) -anchor nw]
	} else {
	    $ds9(canvas) coords $graph(horz,id) $xx $yy
	}
	$ds9(graph,horz) configure -width $ww

    } else {
	if {$graph(horz,id)>0} {
	    $ds9(canvas) delete $graph(horz,id)
	    set graph(horz,id) 0
	}
    }

    if {$view(graph,vert)} {
	set xx [expr $canvas(width)+$graph(gap,x)]
	set yy 0
	set hh [expr $canvas(height)+$graph(gap,y)]

	if {$graph(vert,id) == 0} {
	    set graph(vert,id) [$ds9(canvas) create window $xx $yy \
				    -window $ds9(graph,vert) -anchor nw]
	} else {
	    $ds9(canvas) coords $graph(vert,id) $xx $yy
	}
	$ds9(graph,vert) configure -height $hh

    } else {
	if {$graph(vert,id)>0} {
	    $ds9(canvas) delete $graph(vert,id)
	    set graph(vert,id) 0
	}
    }
}

proc LayoutFrames {} {
    global ds9
    global current
    global tile
    global view
    global colorbar
    
    # turn everything off
    foreach f $ds9(frames) {
	$f hide
	$f highlite off
	$f panner off
	$f magnifier off
	UnBindEventsFrame $f
    }

    set n [llength $ds9(active)]
    if {$n > 0} {
	switch -- $ds9(display,mode) {
	    single {TileOne}
	    tile {
		switch -- $tile(mode) {
		    row {TileRect 1 $n $tile(grid,gap)}
		    column {TileRect $n 1 $tile(grid,gap)}
		    grid {
			switch -- $tile(grid,mode) {
			    automatic {
				TileRect [expr int(sqrt($n-1))+1] \
				    [expr int(sqrt($n)+.5)] \
				    $tile(grid,gap)
			    }
			    manual {
				TileRect $tile(grid,row) \
				    $tile(grid,col) \
				    $tile(grid,gap)
			    }
			}
		    }
		}
	    }
	    blink {TileOne}
	}
    } else {
	set current(frame) {}
	set ds9(next) {}

	# panner
	if {$view(panner)} {
	    panner clear
	}

	# magnifier
	if {$view(magnifier)} {
	    magnifier clear
	}

	# process proper colorbar
	switch -- $ds9(visual) {
	    pseudocolor {
		colorbar show
		set current(colorbar) colorbar
		set colorbar(map) [colorbar get name]
		set colorbar(invert) [colorbar get invert]
	    }
	    truecolor {
		colorbar show
		colorbarrgb hide
		set current(colorbar) colorbar
		set colorbar(map) [colorbar get name]
		set colorbar(invert) [colorbar get invert]
	    }
	}

	# update menus/dialogs
	UpdateDS9
    }
}

proc LayoutInfoPanel {} {
    global ds9
    global view
    global info

    global debug
    if {$debug(tcl,layout)} {
	puts "LayoutInfoPanel"
    }

    # reset
    grid columnconfigure $ds9(info) 5 -weight 0
    if {$info(row) > -1} {
	grid rowconfigure $ds9(info) $info(row) -weight 0
	set info(row) -1
    }

    # layout
    switch $view(layout) {
	vertical {LayoutInfoPanelVert}
	horizontal -
	default {LayoutInfoPanelHorz}
    }
}

proc LayoutInfoPanelHorz {} {
    global ds9
    global view

    set ww 13
    set xx 31
    set row 0

    grid columnconfigure $ds9(info) 5 -weight 1

    # filename
    if {$view(info,filename)} {
	$ds9(info).fileValue configure -width $xx

	grid $ds9(info).fileTitle -row $row -column 0 -sticky w
	grid $ds9(info).fileValue -row $row -column 2 -padx 2 \
	    -sticky ew -columnspan 4
	incr row
    } else {
	grid forget $ds9(info).fileTitle
	grid forget $ds9(info).fileValue
    }

    # object
    if {$view(info,object)} {
	$ds9(info).objValue configure -width $xx

	grid $ds9(info).objTitle -row $row -column 0 -sticky w
	grid $ds9(info).objValue -row $row -column 2 -padx 2 \
	    -sticky ew -columnspan 4
	incr row
    } else {
	grid forget $ds9(info).objTitle
	grid forget $ds9(info).objValue
    }

    # minmax
    if {$view(info,minmax)} {
	$ds9(info).minValue configure -width $ww
	$ds9(info).maxValue configure -width $ww

	grid $ds9(info).minTitle -row $row -column 0 -sticky w
	grid $ds9(info).minValue -row $row -column 2 -padx 2 -sticky w
	grid $ds9(info).maxTitle -row $row -column 3 -sticky w
	grid $ds9(info).maxValue -row $row -column 4 -padx 2 -sticky w
	incr row
    } else {
	grid forget $ds9(info).minTitle
	grid forget $ds9(info).minValue
	grid forget $ds9(info).maxTitle
	grid forget $ds9(info).maxValue
    }

    # lowhigh
    if {$view(info,lowhigh)} {
	$ds9(info).lowValue configure -width $ww
	$ds9(info).highValue configure -width $ww

	grid $ds9(info).lowTitle -row $row -column 0 -sticky w
	grid $ds9(info).lowValue -row $row -column 2 -padx 2 -sticky w
	grid $ds9(info).highTitle -row $row -column 3 -sticky w
	grid $ds9(info).highValue -row $row -column 4 -padx 2 -sticky w
	incr row
    } else {
	grid forget $ds9(info).lowTitle
	grid forget $ds9(info).lowValue
	grid forget $ds9(info).highTitle
	grid forget $ds9(info).highValue
    }

    # value
    $ds9(info).valueValue.v configure -width $ww
    $ds9(info).valueValue.red configure -width [expr $xx/3]
    $ds9(info).valueValue.green configure -width [expr $xx/3]
    $ds9(info).valueValue.blue configure -width [expr $xx/3-1]

    grid $ds9(info).valueTitle -row $row -column 0 -sticky w
    grid $ds9(info).valueValue -row $row -column 2 -padx 2 \
	-sticky w -columnspan 3

    grid $ds9(info).valueValue.v
    grid forget $ds9(info).valueValue.red
    grid forget $ds9(info).valueValue.green
    grid forget $ds9(info).valueValue.blue
    incr row

    # wcs
    foreach l {{} a b c d e f g h i j k l m n o p q r s t u v w x y z} {
	$ds9(info).wcsXValue$l configure -width $ww
	$ds9(info).wcsYValue$l configure -width $ww

	if {$view(info,wcs$l)} {
	    grid $ds9(info).wcsLabel$l  -row $row -column 0 -sticky w
	    grid $ds9(info).wcsXLabel$l -row $row -column 1 -sticky e
	    grid $ds9(info).wcsXValue$l -row $row -column 2 -padx 2
	    grid $ds9(info).wcsYLabel$l -row $row -column 3 -sticky e

	    grid $ds9(info).wcsYValue$l -row $row -column 4 -padx 2
	    incr row
	} else {
	    grid forget $ds9(info).wcsLabel$l
	    grid forget $ds9(info).wcsXLabel$l
	    grid forget $ds9(info).wcsXValue$l
	    grid forget $ds9(info).wcsYLabel$l
	    grid forget $ds9(info).wcsYValue$l
	}
    }

    # detector
    if {$view(info,detector)} {
	$ds9(info).detectorXValue configure -width $ww
	$ds9(info).detectorYValue configure -width $ww

	grid $ds9(info).detectorTitle -row $row -column 0 -sticky w
	grid $ds9(info).detectorXLabel -row $row -column 1 -sticky e
	grid $ds9(info).detectorXValue -row $row -column 2 -padx 2
	grid $ds9(info).detectorYLabel -row $row -column 3 -sticky e
	grid $ds9(info).detectorYValue -row $row -column 4 -padx 2
	incr row
    } else {
	grid forget $ds9(info).detectorTitle
	grid forget $ds9(info).detectorXLabel
	grid forget $ds9(info).detectorXValue
	grid forget $ds9(info).detectorYLabel
	grid forget $ds9(info).detectorYValue
    }

    # amplifier
    if {$view(info,amplifier)} {
	$ds9(info).amplifierXValue configure -width $ww
	$ds9(info).amplifierYValue configure -width $ww

	grid $ds9(info).amplifierTitle -row $row -column 0 -sticky w
	grid $ds9(info).amplifierXLabel -row $row -column 1 -sticky e
	grid $ds9(info).amplifierXValue -row $row -column 2 -padx 2
	grid $ds9(info).amplifierYLabel -row $row -column 3 -sticky e
	grid $ds9(info).amplifierYValue -row $row -column 4 -padx 2
	incr row
    } else {
	grid forget $ds9(info).amplifierTitle
	grid forget $ds9(info).amplifierXLabel
	grid forget $ds9(info).amplifierXValue
	grid forget $ds9(info).amplifierYLabel
	grid forget $ds9(info).amplifierYValue
    }

    # physical
    if {$view(info,physical)} {
	$ds9(info).physicalXValue configure -width $ww
	$ds9(info).physicalYValue configure -width $ww 

	grid $ds9(info).physicalTitle -row $row -column 0 -sticky w
	grid $ds9(info).physicalXLabel -row $row -column 1 -sticky e
	grid $ds9(info).physicalXValue -row $row -column 2 -padx 2
	grid $ds9(info).physicalYLabel -row $row -column 3 -sticky e
	grid $ds9(info).physicalYValue -row $row -column 4 -padx 2
	incr row
    } else {
	grid forget $ds9(info).physicalTitle
	grid forget $ds9(info).physicalXLabel
	grid forget $ds9(info).physicalXValue
	grid forget $ds9(info).physicalYLabel
	grid forget $ds9(info).physicalYValue
    }

    # image
    if {$view(info,image)} {
	$ds9(info).imageXValue configure -width $ww 
	$ds9(info).imageYValue configure -width $ww

	grid $ds9(info).imageTitle -row $row -column 0 -sticky w
	grid $ds9(info).imageXLabel -row $row -column 1 -sticky e
	grid $ds9(info).imageXValue -row $row -column 2 -padx 2
	grid $ds9(info).imageYLabel -row $row -column 3 -sticky e
	grid $ds9(info).imageYValue -row $row -column 4 -padx 2
	incr row
    } else {
	grid forget $ds9(info).imageTitle
	grid forget $ds9(info).imageXLabel
	grid forget $ds9(info).imageXValue
	grid forget $ds9(info).imageYLabel
	grid forget $ds9(info).imageYValue
    }

    # frame, zoom, angle
    if {$view(info,frame)} {
	$ds9(info).zoomValue configure -width $ww
	$ds9(info).angleValue configure -width $ww

	grid $ds9(info).frame -row $row -column 0 -sticky w
	grid $ds9(info).zoomtitle -row $row -column 1 -sticky w
	grid $ds9(info).zoomValue -row $row -column 2 -padx 2 -sticky w
	grid $ds9(info).angleTitle -row $row -column 3
	grid $ds9(info).angleValue -row $row -column 4 -padx 2 -sticky w
	incr row
    } else {
	grid forget $ds9(info).frame
	grid forget $ds9(info).zoomtitle
	grid forget $ds9(info).zoomValue
	grid forget $ds9(info).angleTitle
	grid forget $ds9(info).angleValue
    }

    # dummy
    grid forget $ds9(info).dummy
}

proc LayoutInfoPanelVert {} {
    global ds9
    global view

    set ww 11
    set row 0

    # filename
    if {$view(info,filename)} {
	$ds9(info).fileValue configure -width $ww

	grid $ds9(info).fileTitle -row $row -column 0 -sticky e 
	grid $ds9(info).fileValue -row $row -column 1 -padx 2 -sticky w
	incr row
    } else {
	grid forget $ds9(info).fileTitle
	grid forget $ds9(info).fileValue
    }

    # object
    if {$view(info,object)} {
	$ds9(info).objValue configure -width $ww

	grid $ds9(info).objTitle -row $row -column 0 -sticky e 
	grid $ds9(info).objValue -row $row -column 1 -padx 2 -sticky w
	incr row
    } else {
	grid forget $ds9(info).objTitle
	grid forget $ds9(info).objValue
    }

    # minmax
    if {$view(info,minmax)} {
	$ds9(info).minValue configure -width $ww
	$ds9(info).maxValue configure -width $ww

	grid $ds9(info).minTitle -row $row -column 0 -sticky e
	grid $ds9(info).minValue -row $row -column 1 -padx 2 -sticky w
	incr row
	grid $ds9(info).maxTitle -row $row -column 0 -sticky e
	grid $ds9(info).maxValue -row $row -column 1 -padx 2 -sticky w
	incr row
    } else {
	grid forget $ds9(info).minTitle
	grid forget $ds9(info).minValue
	grid forget $ds9(info).maxTitle
	grid forget $ds9(info).maxValue
    }

    # lowhigh
    if {$view(info,lowhigh)} {
	$ds9(info).lowValue configure -width $ww
	$ds9(info).highValue configure -width $ww

	grid $ds9(info).lowTitle -row $row -column 0 -sticky e
	grid $ds9(info).lowValue -row $row -column 1 -padx 2 -sticky w
	incr row
	grid $ds9(info).highTitle -row $row -column 0 -sticky e
	grid $ds9(info).highValue -row $row -column 1 -padx 2 -sticky w
	incr row
    } else {
	grid forget $ds9(info).lowTitle
	grid forget $ds9(info).lowValue
	grid forget $ds9(info).highTitle
	grid forget $ds9(info).highValue
    }

    # value
    $ds9(info).valueValue.v configure -width $ww
    $ds9(info).valueValue.red configure -width [expr $ww/3]
    $ds9(info).valueValue.green configure -width [expr $ww/3]
    $ds9(info).valueValue.blue configure -width [expr $ww/3]

    grid $ds9(info).valueTitle -row $row -column 0 -sticky e
    grid $ds9(info).valueValue -row $row -column 1 -padx 2 -sticky w

    grid $ds9(info).valueValue.v
    grid forget $ds9(info).valueValue.red
    grid forget $ds9(info).valueValue.green
    grid forget $ds9(info).valueValue.blue
    incr row

    # wcs
    foreach l {{} a b c d e f g h i j k l m n o p q r s t u v w x y z} {
	$ds9(info).wcsXValue$l configure -width $ww
	$ds9(info).wcsYValue$l configure -width $ww

	if {$view(info,wcs$l)} {
	    grid $ds9(info).wcsLabel$l  -row $row -column 1 -sticky ew
	    incr row
	    grid $ds9(info).wcsXLabel$l -row $row -column 0 -sticky e
	    grid $ds9(info).wcsXValue$l -row $row -column 1 -padx 2
	    incr row
	    grid $ds9(info).wcsYLabel$l -row $row -column 0 -sticky e
	    grid $ds9(info).wcsYValue$l -row $row -column 1 -padx 2
	    incr row
	} else {
	    grid forget $ds9(info).wcsLabel$l
	    grid forget $ds9(info).wcsXLabel$l
	    grid forget $ds9(info).wcsXValue$l
	    grid forget $ds9(info).wcsYLabel$l
	    grid forget $ds9(info).wcsYValue$l
	}
    }

    # detector
    if {$view(info,detector)} {
	$ds9(info).detectorXValue configure -width $ww
	$ds9(info).detectorYValue configure -width $ww

	grid $ds9(info).detectorTitle -row $row -column 1 -sticky ew
	incr row
	grid $ds9(info).detectorXLabel -row $row -column 0 -sticky e
	grid $ds9(info).detectorXValue -row $row -column 1 -padx 2
	incr row
	grid $ds9(info).detectorYLabel -row $row -column 0 -sticky e
	grid $ds9(info).detectorYValue -row $row -column 1 -padx 2
	incr row
    } else {
	grid forget $ds9(info).detectorTitle
	grid forget $ds9(info).detectorXLabel
	grid forget $ds9(info).detectorXValue
	grid forget $ds9(info).detectorYLabel
	grid forget $ds9(info).detectorYValue
    }

    # amplifier
    if {$view(info,amplifier)} {
	$ds9(info).amplifierXValue configure -width $ww
	$ds9(info).amplifierYValue configure -width $ww

	grid $ds9(info).amplifierTitle -row $row -column 1 -sticky ew
	incr row
	grid $ds9(info).amplifierXLabel -row $row -column 0 -sticky e
	grid $ds9(info).amplifierXValue -row $row -column 1 -padx 2
	incr row
	grid $ds9(info).amplifierYLabel -row $row -column 0 -sticky e
	grid $ds9(info).amplifierYValue -row $row -column 1 -padx 2
	incr row
    } else {
	grid forget $ds9(info).amplifierTitle
	grid forget $ds9(info).amplifierXLabel
	grid forget $ds9(info).amplifierXValue
	grid forget $ds9(info).amplifierYLabel
	grid forget $ds9(info).amplifierYValue
    }

    # physical
    if {$view(info,physical)} {
	$ds9(info).physicalXValue configure -width $ww
	$ds9(info).physicalYValue configure -width $ww 

	grid $ds9(info).physicalTitle -row $row -column 1 -sticky ew
	incr row
	grid $ds9(info).physicalXLabel -row $row -column 0 -sticky e
	grid $ds9(info).physicalXValue -row $row -column 1 -padx 2
	incr row
	grid $ds9(info).physicalYLabel -row $row -column 0 -sticky e
	grid $ds9(info).physicalYValue -row $row -column 1 -padx 2
	incr row
    } else {
	grid forget $ds9(info).physicalTitle
	grid forget $ds9(info).physicalXLabel
	grid forget $ds9(info).physicalXValue
	grid forget $ds9(info).physicalYLabel
	grid forget $ds9(info).physicalYValue
    }

    # image
    if {$view(info,image)} {
	$ds9(info).imageXValue configure -width $ww 
	$ds9(info).imageYValue configure -width $ww

	grid $ds9(info).imageTitle -row $row -column 1 -sticky ew
	incr row
	grid $ds9(info).imageXLabel -row $row -column 0 -sticky e
	grid $ds9(info).imageXValue -row $row -column 1 -padx 2
	incr row
	grid $ds9(info).imageYLabel -row $row -column 0 -sticky e
	grid $ds9(info).imageYValue -row $row -column 1 -padx 2
	incr row
    } else {
	grid forget $ds9(info).imageTitle
	grid forget $ds9(info).imageXLabel
	grid forget $ds9(info).imageXValue
	grid forget $ds9(info).imageYLabel
	grid forget $ds9(info).imageYValue
    }

    # frame, zoom, angle
    if {$view(info,frame)} {
	$ds9(info).zoomValue configure -width $ww
	$ds9(info).angleValue configure -width $ww

	grid $ds9(info).frame -row $row -column 1 -sticky ew
	incr row
	grid $ds9(info).zoomtitle -row $row -column 0 -sticky e
	grid $ds9(info).zoomValue -row $row -column 1 -padx 2
	incr row
	grid $ds9(info).angleTitle -row $row -column 0 -sticky e
	grid $ds9(info).angleValue -row $row -column 1 -padx 2
	incr row
    } else {
	grid forget $ds9(info).frame
	grid forget $ds9(info).zoomtitle
	grid forget $ds9(info).zoomValue
	grid forget $ds9(info).angleTitle
	grid forget $ds9(info).angleValue
    }

    # dummy
    global info
    set info(row) $row

    grid $ds9(info).dummy -row $row -column 1
    grid rowconfigure $ds9(info) $row -weight 1
}

proc LayoutInfoPanelValue {which} {
    global ds9

    switch -- [$which get type] {
	base {
	    grid forget $ds9(info).valueValue.red
	    grid forget $ds9(info).valueValue.green
	    grid forget $ds9(info).valueValue.blue
	    grid $ds9(info).valueValue.v
	}
	rgb {
	    grid forget $ds9(info).valueValue.v
	    grid $ds9(info).valueValue.red \
		$ds9(info).valueValue.green \
		$ds9(info).valueValue.blue
	}
    }
}

# This procedure is called when we have only 1 frames to display

proc TileOne {} {
    global ds9
    global view
    global current
    global canvas
    global colorbar

    global debug
    if {$debug(tcl,layout)} {
	puts "TileOne"
    }

    set ww $canvas(width)
    set hh $canvas(height)

    set xx 0
    set yy 0

    foreach f $ds9(active) {
	$f configure -x $xx -y $yy -width $ww -height $hh -anchor nw
    }

    # only show the current frame
    $current(frame) show
    FrameToFront $current(frame)
}

proc TileRect {numx numy gap} {
    global view
    global canvas

    global debug
    if {$debug(tcl,layout)} {
	puts "TileRect"
    }

    set ww $canvas(width)
    set hh $canvas(height)

    set w [expr int(($ww-$gap*($numx-1))/$numx)]
    set h [expr int(($hh-$gap*($numy-1))/$numy)]
    
    for {set jj 0} {$jj<$numy} {incr jj} {
	for {set ii 0} {$ii<$numx} {incr ii} {
	    set c [expr $jj*$numx + $ii]
	    set x($c) [expr ($w+$gap)*$ii]
	    set y($c) [expr ($h+$gap)*$jj]
	}
    }

    TileIt $w $h x y [expr $numx*$numy]
}

proc TileIt {ww hh xx yy nn} {
    upvar $xx x
    upvar $yy y
    global ds9
    global current

    global debug
    if {$debug(tcl,layout)} {
	puts "TileIt"
    }

    set i 0
    foreach f $ds9(active) {
	if {$i<$nn} {
	    $f configure -x $x($i) -y $y($i) -width $ww -height $hh -anchor nw
	    $f show
	    $ds9(canvas) raise $f
	    if {!$ds9(freeze)} {
		BindEventsFrame $f
	    }
	}
	incr i
    }

    # if manual grid, current frame could be not included
    if {[llength $ds9(active)] > $nn} {
	set current(frame) [lindex $ds9(active) 0]
    }
    FrameToFront $current(frame)
}

proc DisplayDefaultDialog {} {
    global ed
    global canvas

    set w {.defdpy}

    set ed(ok) 0
    set ed(x) $canvas(width)
    set ed(y) $canvas(height)

    DialogCreate $w [msgcat::mc {Display Size}]

    frame $w.ed -relief ridge -bd 1
    frame $w.buttons -relief ridge -bd 1
    pack $w.ed $w.buttons -fill x -ipadx 4 -ipady 4

    label $w.ed.xTitle -text {X}
    label $w.ed.yTitle -text {Y}
    entry $w.ed.x -textvariable ed(x) -width 10
    entry $w.ed.y -textvariable ed(y) -width 10
    label $w.ed.xunit -text [msgcat::mc {Pixels}]
    label $w.ed.yunit -text [msgcat::mc {Pixels}]
    
    grid $w.ed.xTitle $w.ed.x $w.ed.xunit -padx 4 -sticky w
    grid $w.ed.yTitle $w.ed.y $w.ed.yunit -padx 4 -sticky w

    button $w.buttons.ok -text [msgcat::mc {OK}]  \
	-default active -command {set ed(ok) 1}
    button $w.buttons.cancel -text [msgcat::mc {Cancel}] \
	-command {set ed(ok) 0}
    pack $w.buttons.ok -side left -padx 10
    pack $w.buttons.cancel -side right -padx 10

    bind $w <Return> {set ed(ok) 1}
    bind $w <Alt-o> "tkButtonInvoke $w.buttons.ok"
    bind $w <Alt-c> "tkButtonInvoke $w.buttons.cancel"

    DialogCenter $w 
    $w.ed.x select range 0 end
    DialogWait $w ed(ok) $w.ed.x
    DialogDismiss $w

    if {$ed(ok)} {
	set canvas(width) $ed(x)
	set canvas(height) $ed(y)
	LayoutDS9
    }

    unset ed
}

# Process Cmds

proc ProcessWidthCmd {varname iname} {
    upvar $varname var
    upvar $iname i

    # we must have a window to change
    RealizeDS9

    global canvas
    set canvas(width) [lindex $var $i]
    LayoutDS9
}

proc ProcessHeightCmd {varname iname} {
    upvar $varname var
    upvar $iname i

    # we must have a window to change
    RealizeDS9

    global canvas
    set canvas(height) [lindex $var $i]
    LayoutDS9
}

proc ProcessViewCmd {varname iname} {
    upvar $varname var
    upvar $iname i

    global view
    global rgb

    set item [string tolower [lindex $var $i]]

    switch  -- $item {
	horizontal -
	vertical {LayoutViewReset; set view(layout) $item}

	default {
	    set yesno [lindex $var [expr $i+1]]

	    switch -- $item {
		info {set view(info) [FromYesNo $yesno]}
		panner {set view(panner) [FromYesNo $yesno]}
		magnifier {set view(magnifier) [FromYesNo $yesno]}
		buttons {set view(buttons) [FromYesNo $yesno]}
		colorbar {set view(colorbar) [FromYesNo $yesno]}
		colorbarnumerics {set view(colorbar,numerics) [FromYesNo $yesno]}
		horzgraph {set view(graph,horz) [FromYesNo $yesno]}
		vertgraph {set view(graph,vert) [FromYesNo $yesno]}

		filename {set view(info,filename) [FromYesNo $yesno]}
		object {set view(info,object) [FromYesNo $yesno]}
		minmax {set view(info,minmax) [FromYesNo $yesno]}
		lowhigh {set view(info,lowhigh) [FromYesNo $yesno]}
		frame {set view(info,frame) [FromYesNo $yesno]}

		detector {set view(info,detector) [FromYesNo $yesno]}
		amplifier {set view(info,amplifier) [FromYesNo $yesno]}
		physical {set view(info,physical) [FromYesNo $yesno]}
		image {set view(info,image) [FromYesNo $yesno]}
		wcs {set view(info,wcs) [FromYesNo $yesno]}
		wcsa {set view(info,wcsa) [FromYesNo $yesno]}
		wcsb {set view(info,wcsb) [FromYesNo $yesno]}
		wcsc {set view(info,wcsc) [FromYesNo $yesno]}
		wcsd {set view(info,wcsd) [FromYesNo $yesno]}
		wcse {set view(info,wcse) [FromYesNo $yesno]}
		wcsf {set view(info,wcsf) [FromYesNo $yesno]}
		wcsg {set view(info,wcsg) [FromYesNo $yesno]}
		wcsh {set view(info,wcsh) [FromYesNo $yesno]}
		wcsi {set view(info,wcsi) [FromYesNo $yesno]}
		wcsj {set view(info,wcsj) [FromYesNo $yesno]}
		wcsk {set view(info,wcsk) [FromYesNo $yesno]}
		wcsl {set view(info,wcsl) [FromYesNo $yesno]}
		wcsm {set view(info,wcsm) [FromYesNo $yesno]}
		wcsn {set view(info,wcsn) [FromYesNo $yesno]}
		wcso {set view(info,wcso) [FromYesNo $yesno]}
		wcsp {set view(info,wcsp) [FromYesNo $yesno]}
		wcsq {set view(info,wcsq) [FromYesNo $yesno]}
		wcsr {set view(info,wcsr) [FromYesNo $yesno]}
		wcss {set view(info,wcss) [FromYesNo $yesno]}
		wcst {set view(info,wcst) [FromYesNo $yesno]}
		wcsu {set view(info,wcsu) [FromYesNo $yesno]}
		wcsv {set view(info,wcsv) [FromYesNo $yesno]}
		wcsw {set view(info,wcsw) [FromYesNo $yesno]}
		wcsx {set view(info,wcsx) [FromYesNo $yesno]}
		wcsy {set view(info,wcsy) [FromYesNo $yesno]}
		wcsz {set view(info,wcsz) [FromYesNo $yesno]}

		red {set rgb(red) [FromYesNo $yesno]; RGBView}
		green {set rgb(green) [FromYesNo $yesno]; RGBView}
		blue {set rgb(blue) [FromYesNo $yesno]; RGBView}
	    }
	    incr i
	}
    }
    UpdateView
}

