#  Copyright (C) 1999-2008
#  Smithsonian Astrophysical Observatory, Cambridge, MA, USA
#  For conditions of distribution and use, see copyright notice in "copyright"

package provide DS9 1.0

proc NSVRServer {varname} {
    upvar #0 $varname var
    global $varname

    global debug
    if {$debug(tcl,image)} {
	puts "NSVRServer $varname"
    }

    global nres

    ARStatus $varname "Looking up $var(name)"

    switch -- $nres(server) {
	ned-sao {
	    set var(query) "[regsub -all {\ } $var(name) {%20}]"
	    set var(url) "http://www.cfa.harvard.edu/catalog/ned"
	    set var(cmd) NSVRNEDSAO
	}
	ned-eso {
	    set var(query) "&o=[regsub -all {\ } $var(name) {%20}]"
	    set var(url) "http://archive.eso.org/skycat/servers/ned-server"
	    set var(cmd) NSVRNEDESO
	}
	simbad-sao {
	    set var(query) "[regsub -all {\ } $var(name) {%20}]"
	    set var(url) "http://www.cfa.harvard.edu/catalog/simbad-cfa"
	    set var(cmd) NSVRSIMBADSAO
	}
	simbad-eso {
	    set var(query) "&o=[regsub -all {\ } $var(name) {%20}]"
	    set var(url) "http://archive.eso.org/skycat/servers/sim-server"
	    set var(cmd) NSVRSIMBADESO
	}
    }

    NSVRLoad $varname
}

proc NSVRLoad {varname} {
    upvar #0 $varname var
    global $varname

    global debug
    if {$debug(tcl,image)} {
	puts "NSVRLoad $varname $var(url) $var(query)"
    }

    global http

    # -query will not work, do it manually
    if {$var(sync)} {
	if {![catch {set var(token) [http::geturl $var(url)?$var(query) \
					 -timeout $http(timeout) \
					 -headers "[ProxyHTTP]"]
	}]} {
	    set var(active) 1
	    NSVRLoadFinish $varname $var(token)
	} else {
	    ARError $varname "[msgcat::mc {Unable to locate URL}] $var(url)"
	}
    } else {
	if {![catch {set var(token) [http::geturl $var(url)?$var(query) \
					 -timeout $http(timeout) \
					 -command \
					 [list NSVRLoadFinish $varname] \
					 -headers "[ProxyHTTP]"]
	}]} {
	    set var(active) 1
	} else {
	    ARError $varname "[msgcat::mc {Unable to locate URL}] $var(url)"
	}
    }
}

proc NSVRLoadFinish {varname token} {
    upvar #0 $varname var
    global $varname

    global debug
    if {$debug(tcl,image)} {
	puts "NSVRLoadFinish $varname"
    }

    if {!($var(active))} {
	ARCancelled $varname
	return
    }

    upvar #0 $token t

    # Code
    set code [http::ncode $token]

    # Meta
    set meta $t(meta)

    # Log it
    HTTPLog $token

    # Result?
    switch -- $code {
	200 -
	203 -
	404 -
	503 {eval $var(cmd) $varname}

	201 -
	300 -
	301 -
	302 -
	303 -
	305 -
	307 {
	    foreach {name value} $meta {
		if {[regexp -nocase ^location$ $name]} {
		    global debug
		    if {$debug(tcl,image)} {
			puts "NSVRLoadFinish redirect $code to $value"
		    }
		    # clean up and resubmit
		    http::cleanup $token
		    unset var(token)

		    # strip query from url
		    set var(url) [lindex [split $value {?}] 0]

		    NSVRLoad $varname
		}
	    }
	}

	default {ARError $varname "[msgcat::mc {Error code was returned}] $code"}
    }
}

proc NSVRNEDSAO {varname} {
    upvar #0 $varname var
    global $varname

    # data is in sexagesmal
    set data [http::data $var(token)]
    set r [lindex $data 0]
    set d  [lindex $data 1]
    if {($r != {}) && ($d != {})} {
	switch -- $var(skyformat) {
	    degrees {NSVRParse $varname [h2d [Sex2H $r]] [Sex2D $d]}
	    sexagesimal {NSVRParse $varname $r $d}
	}
    } else {
	NSVRParse $varname {} {}
    }
}

proc NSVRNEDESO {varname} {
    upvar #0 $varname var
    global $varname

    # data is in degrees
    set data [http::data $var(token)]
    set l [llength $data]
    set r [lindex $data [expr $l-3]]
    set d [lindex $data [expr $l-2]]
    if {[string is double -strict $r] && [string is double -strict $d]} {
	switch -- $var(skyformat) {
	    degrees {NSVRParse $varname $r $d}
	    sexagesimal {
		NSVRParse $varname [uformat d h: $r] [uformat d d: $d]
	    }
	}
    } else {
	NSVRParse $varname {} {}
    }
}

proc NSVRSIMBADSAO {varname} {
    upvar #0 $varname var
    global $varname

    # data is in sexagesmal
    set data [http::data $var(token)]
    set r [lindex $data 0]
    set d  [lindex $data 1]
    if {($r != {}) && ($d != {})} {
	switch -- $var(skyformat) {
	    degrees {NSVRParse $varname [h2d [Sex2H $r]] [Sex2D $d]}
	    sexagesimal {NSVRParse $varname $r $d}
	}
    } else {
	NSVRParse $varname {} {}
    }
}

proc NSVRSIMBADESO {varname} {
    upvar #0 $varname var
    global $varname

    # data is in degrees
    set data [http::data $var(token)]
    set l [llength $data]
    set r [lindex $data [expr $l-3]]
    set d [lindex $data [expr $l-2]]
    if {[string is double -strict $r] && [string is double -strict $d]} {
	switch -- $var(skyformat) {
	    degrees {NSVRParse $varname $r $d}
	    sexagesimal {
		NSVRParse $varname [uformat d h: $r] [uformat d d: $d]
	    }
	}
    } else {
	NSVRParse $varname {} {}
    }
}

proc NSVRParse {varname x y} {
    upvar #0 $varname var
    global $varname

    global debug
    if {$debug(tcl,image)} {
	puts "NSVRParse $varname"
    }

    set var(x) $x
    set var(y) $y

    if {($var(x) == {}) || ($var(y) == {})} {
	set var(x) {}
	set var(y) {}

	ARStatus $varname "$var(name) not found"
	ARReset $varname
    } else {
	if {$var(proc,next) != {}} {
	    http::cleanup $var(token)
	    unset var(token)

	    eval $var(proc,next) $varname
	} else {
	    ARDone $varname
	}
    }
}

proc NSVRServerMenu {varname} {
    upvar #0 $varname var
    global $varname

    global nres

    $var(mb) add cascade -label [msgcat::mc {Name Server}] -menu $var(mb).name
    menu $var(mb).name
    $var(mb).name add radiobutton -label {NED@SAO} \
	-variable nres(server) -value ned-sao
    $var(mb).name add radiobutton -label {NED@ESO} \
	-variable nres(server) -value ned-eso
    $var(mb).name add radiobutton -label {SIMBAD@SAO} \
	-variable nres(server) -value simbad-sao
    $var(mb).name add radiobutton -label {SIMBAD@ESO} \
	-variable nres(server) -value simbad-eso
}
