#  Copyright (C) 1999-2008
#  Smithsonian Astrophysical Observatory, Cambridge, MA, USA
#  For conditions of distribution and use, see copyright notice in "copyright"

package provide DS9 1.0

proc UpdateDS9 {} {
    global debug
    if {$debug(tcl,update)} {
	puts ""
	puts "UpdateDS9 begin..."
    }

    UpdateFileMenu
    UpdateEditMenu
    UpdateFrameMenu
    UpdateCubeDialog
    UpdateRGBDialog
    UpdateBinMenu
    UpdateBinDialog
    UpdateZoomMenu
    UpdateZoomDialog
    UpdateScaleMenu
    UpdateScaleDialog
    UpdateColorMenu
    UpdateColorDialog
    UpdateColormapLevel
    UpdateRegionMenu
    UpdateGroupDialog
    UpdateWCSMenu
    UpdateAnalysisMenu

    UpdateMaskMenu
    UpdateContourMenu
    UpdateContourDialog
    UpdateGridMenu
    UpdateGridDialog
    UpdateSmoothMenu
    UpdateSmoothDialog

    UpdateGraphYAxis
    UpdateGraphXAxis

    UpdateInfoBoxCurrent
    
    if {$debug(tcl,update)} {
	puts "UpdateDS9 end...\n"
    }
}

# capture general errors
# this only captures gui errors, not xpa errors
proc bgerror {err} {
    global cursor

    tk_messageBox -type ok -icon error \
	-message "[msgcat::mc {An internal error has been detected}] $err"

    # reset cursor
    set cursor(count) 0
    UnsetWatchCursor
}

# force capture xpa/interactive errors
proc Error {message} {
    global ds9

    if {$ds9(xpa,ptr) == 0} {
	tk_messageBox -message $message -type ok -icon error
    } else {
	xpaerror $ds9(xpa,ptr) $message
    }

    # reset cursor
    set cursor(count) 0
    UnsetWatchCursor
}

proc Info {message} {
    global ds9

    if {$ds9(xpa,ptr) == 0} {
	tk_messageBox -message $message -type ok -icon info
    } else {
	xpaerror $ds9(xpa,ptr) $message
    }

    # reset cursor
    set cursor(count) 0
    UnsetWatchCursor
}

proc SourceInitFile {ext} {
    global ds9
    global env
    
    foreach ff {{.} {}} {
	set fn $ff$ds9(app)$ext
	foreach dir [list {.} [GetEnvHome] {/usr/local/lib} {/opt/local/lib}] {
	    set ff [file join $dir $fn]
	    if [file exist $ff] {
		# can't make this a debug command line option
		# prefs set before options parsed
		if [catch {source $ff}] {
		    Error "[msgcat::mc {An error has occured while executing}] $ff. [msgcat::mc {DS9 will complete the initialization process}]"
		}
		return
	    }
	}
    }
}

proc DeleteInitFile {ext} {
    global ds9
    global env
    
    foreach ff {{.} {}} {
	set fn $ff$ds9(app)$ext
	foreach dir [list {.} [GetEnvHome]] {
	    set ff [file join $dir $fn]
	    if [file exist $ff] {
		catch {file delete -force $ff}
		return
	    }
	}
    }
}

proc LanguageToName {which} {
    global ds9

    switch $which {
	locale {return {Locale}}
	da {return {Dansk}}
	de {return {Deutsch}}
	en {return {English}}
	es {return [encoding convertfrom iso8859-1 {Espaol}]}
	fr {return [encoding convertfrom iso8859-1 {Franais}]}
	ja {return [encoding convertfrom euc-jp "\xc6\xfc\xcb\xdc\xb8\xec"]}
	pt {return [encoding convertfrom iso8859-1 {Portugus}]}
    }
}

proc SetLanguage {ll} {
    global ds9
    
    set ds9(language,name) [LanguageToName $ll]

    set x 0
    msgcat::mclocale $ll

    msgcat::mcload [file join $::tk_library msgs]

    # we need to find if we support this language
    if [msgcat::mcload [file join $ds9(root) msgs]] {
	incr x
    }
    if {$ds9(language,dir) != {}} {
	if [msgcat::mcload $ds9(language,dir)] {
	    incr x
	}
    }

    # if english, always return found
    if {![string compare [string range $ll 0 1] {en}]} {
	incr x
    }

    if {$x} {
	return 1
    } else {
	return 0
    }
}

proc GetEnvHome {} {
    global env

    # return current directory by default
    set home {}

    global tcl_platform
    switch $tcl_platform(platform) {
	unix {
	    if [info exists env(HOME)] {
		set home $env(HOME)
	    }
	}
	windows {
	    if {[info exists env(HOME)]} {
		set hh [file normalize [file nativename $env(HOME)]]
		if [file isdirectory $hh] {
		    set home $hh
		}
	    }
	    # this is just a backup, the above should always work
	    if {$home == {} &&
		[info exists env(HOMEDRIVE)] &&
		[info exists env(HOMEPATH)]} {
		set home "$env(HOMEDRIVE)$env(HOMEPATH)"
	    }
	}
    }

    # if there is a problem, just return current directory
    if {![file isdirectory $home]} {
	set home {.}
    }

    return $home
}

proc InitTempDir {} { 
    global ds9
    global env

    # check environment vars first
    #   windows is very picky as to file name format 
    if [info exists env(TEMP)] {
	set ds9(tmpdir) [file normalize [file nativename $env(TEMP)]]
    } elseif [info exists env(TMP)] {
	set ds9(tmpdir) [file normalize [file nativename $env(TMP)]]
    }

    #   nothing so far, go with defaults
    if {![info exists ds9(tmpdir)]} {
	global tcl_platform
	switch $tcl_platform(platform) {
	    unix {set ds9(tmpdir) "/tmp"}
	    windows {set ds9(tmpdir) "C:/WINDOWS/Temp"}
	}
    }

    #   see if it is valid, else current directory
    if {![file isdirectory $ds9(tmpdir)]} {
	set ds9(tmpdir) {.}
    }
}

proc tmpnam {base ext} {
    global ds9

    for {set ii 0} {$ii<10} {incr ii} {
	set fn "$ds9(tmpdir)/$base[clock clicks]$ext"
	if {![file exists $fn]} {
	    return $fn
	}
    }

    # give up
    return $ds9(tmpdir)/$base$ext
}

proc ToYesNo {value} {
    if {$value == 1} {
	return "yes\n"
    } else {
	return "no\n"
    }
}

proc FromYesNo {value} {
    set v [string tolower $value]

    if {$v == "no" || $v == "false" || $v == "off" || $v == 0} {
	return 0
    } else {
	return 1
    }
}

proc RealizeDS9 {} {
    # this has to come first, to realize the canvas
    global debug
    if {$debug(tcl,idletasks)} {
	puts "RealizeDS9"
    }

# idletasks fails for windows. we need to process all events to make
# sure all windows are realized
#    update idletasks
    update

    # now, update all frames
    global ds9
    foreach f $ds9(frames) {
	$f update now
    }
}

proc Sex2H {str} {
    scan $str "%d:%d:%f" h m s
    return [expr $h+($m/60.)+($s/(60.*60.))]
}

proc Sex2D {str} {
    set d 0
    set arcm 0
    set arcs 0
    scan $str "%d:%f:%f" d arcm arcs

    if {$d >= 0} {
	return [expr $d+($arcm/60.)+($arcs/(60.*60.))]
    } else {
	return [expr $d-($arcm/60.)-($arcs/(60.*60.))]
    }
}

proc SetCursor {cursor} {
    global ds9
    global iis
    global current

    # if init phase, don't change cursor
    if {$ds9(init)} {
	return
    }

    # if iis cursor mode, don't change cursor
    if {$iis(state)} {
	return
    }

    if {($current(cursor) != $cursor)} {
	set current(cursor) $cursor
	if {$cursor != {}} {
	    $ds9(canvas) configure -cursor $cursor
	} else {
	    $ds9(canvas) configure -cursor {}
	}
    }
}

# NOTE
# This is where everything gets done. In particular, <Configure> events are
# processed here. Also, any screen updates are done. In short, if it changes
# the geometry or changes the contents, you need to call Set/UnsetWatchCursor
proc SetWatchCursor {} {
    global ds9
    global cursor

    # if init phase, don't change cursor
    if {$ds9(init)} {
	return
    }

    # if not xpa update, don't change cursor/update
    if {!$ds9(xpa,idletasks)} {
	return
    }

    global debug
    if {$debug(tcl,watch)} {
	puts "SetWatchCursor Start $cursor(count)"
    }

    if {$cursor(count) == 0} {
	set cursor(save) [$ds9(canvas) cget -cursor]
	$ds9(canvas) configure -cursor {}
	$ds9(main) configure -cursor watch

	global debug
	if {$debug(tcl,idletasks)} {
	    puts "SetWatchCursor"
	}
	update idletasks
    }
    incr cursor(count)

    if {$debug(tcl,watch)} {
	puts "SetWatchCursor End $cursor(count)"
    }
}

proc UnsetWatchCursor {} {
    global ds9
    global cursor

    # if init phase, don't change cursor
    if {$ds9(init)} {
	return
    }

    # if not xpa update, don't change cursor/update
    if {!$ds9(xpa,idletasks)} {
	return
    }

    global debug
    if {$debug(tcl,watch)} {
	puts "UnsetWatchCursor Start $cursor(count)"
    }

    if {$cursor(count)>0} {
	incr cursor(count) -1
    }

    if {$cursor(count) == 0} {
	$ds9(main) configure -cursor {}
	$ds9(canvas) configure -cursor $cursor(save)

	global debug
	if {$debug(tcl,idletasks)} {
	    puts "UnsetWatchCursor"
	}
	update idletasks
    }    

    if {$debug(tcl,watch)} {
	puts "UnsetWatchCursor End $cursor(count)"
    }
}

proc CursorTimer {} {
    global ds9
    global cursor

    switch -- $cursor(timer) {
	0 {
	    set cursor(timer,abort) 0
	    set cursor(timer) 0
	    set cursor(id) 0
	    $ds9(canvas) configure -cursor {}
	}
	1 {
	    $ds9(canvas) configure -cursor circle
	    set cursor(timer) 2
	    set cursor(id) [after 1000 CursorTimer]
	}
	2 {
	    $ds9(canvas) configure -cursor dot
	    set cursor(timer) 1
	    set cursor(id) [after 1000 CursorTimer]
	}
    }
}

proc AboutBox {} {
    global help
    global ds9
    global about

    set w {.about}
    set width 400
    set height 475

    set about(ok) 0

    DialogCreate $w [msgcat::mc {About SAOImage DS9}]

    frame $w.c -relief ridge -bd 1
    frame $w.buttons -relief ridge -bd 1
    pack $w.c $w.buttons -fill x -ipadx 4 -ipady 4

    canvas $w.c.c -background white -height $height -width $width
    pack $w.c.c
    
    # can't use -file for zvfs
    # set about(sun) [image create photo -format gif -file $ds9(root)/doc/sun.gif]
    set ch [open $ds9(root)/doc/sun.gif r]
    fconfigure $ch -translation binary -encoding binary
    set dd [read $ch]
    close $ch
    unset ch

    set about(sun) [image create photo -format gif -data "$dd"]
    unset dd

    $w.c.c create image 0 0 -image $about(sun) -anchor nw
    $w.c.c create text 120 20 -text $help(about) -anchor nw

    button $w.buttons.ok -text [msgcat::mc {OK}] \
	-default active -command {set about(ok) 1}
    grid $w.buttons.ok -padx 10

    bind $w <Return> {set about(ok) 1}

    DialogCenter $w
    DialogWait $w about(ok)
    DialogDismiss $w

    if {$about(ok)} {
	image delete $about(sun)
    }

    unset about
}

proc QuitDS9 {} {
    catch {IISClose}
    focus {}
    exit
}

proc OpenSource {} {
    set filename [OpenFileDialog tclfbox]
    if {$filename != {}} {
	uplevel #0 "source \{$filename\}"
    }
}

proc OpenConsole {} {
    global ds9

    if {[winfo exists ".tkcon"]} {
	tkcon show
    } else {
	set ::tkcon::OPT(exec) {}
	tkcon::Init
    }
}

proc ToggleBindEvents {} {
    global ds9

    if {$ds9(freeze)} {
	set ds9(freeze) 0
	BindEventsCanvas
	BindEventsPanner
    } else {
	set ds9(freeze) 1
	UnBindEventsCanvas
	UnBindEventsPanner
    }
}

proc ChangeMode {} {
    global ds9

    bind $ds9(canvas) <Button-1> {}
    bind $ds9(canvas) <B1-Motion> {}
    bind $ds9(canvas) <ButtonRelease-1> {}

    foreach f $ds9(frames) {
	$f crosshair off
    }
    UpdateRegionMenu

    foreach f $ds9(frames) {
	$f marker unselect all
	$f marker unhighlite all
	$f marker catalog unselect all
	$f marker catalog unhighlite all
    }

    ClearInfoBox
    PixelTableClearDialog
    ClearGraphData

    switch -- $ds9(mode) {
	none -
	pointer {SetCursor {}}
	crosshair {
	    foreach f $ds9(frames) {
		$f crosshair on
	    }
	    ShowGraphData
	    SetCursor crosshair
	}
	colorbar {
	    bind $ds9(canvas) <Button-1> {ButtonColormap %x %y}
	    bind $ds9(canvas) <B1-Motion> {MotionColormap %x %y}
	    bind $ds9(canvas) <ButtonRelease-1> {ReleaseColormap %x %y}
	    SetCursor center_ptr
	}
	zoom {SetCursor sizing}
	pan {SetCursor fleur}
	rotate {SetCursor exchange}
	examine {SetCursor target}
	illustrate {SetCursor arrow}
	catalog {SetCursor {}}
	imexam {}
    }
}

proc DisplayLog {item} {
    SimpleTextDialog ftptxt [msgcat::mc {Message Log}] 80 40 append bottom $item
}

proc ParseURL {url varname} {
    upvar $varname r

    global ds9

    set r(scheme) {}
    set r(authority) {}
    set r(path) {}
    set r(query) {}
    set r(fragment) {}
    set exp {^(([^:/?#]+):)?(//([^/?#]*))?([^?#]*)(\?([^#]*))?(#(.*))?}

    if {![regexp -nocase $exp $url x a r(scheme) c r(authority) r(path) f r(query) h r(fragment)]} {
	return 0
    }

    # check for windows disk drives
    global tcl_platform
    switch $tcl_platform(platform) {
	unix {}
	windows {
	    switch -- $r(scheme) {
		{} -
		ftp -
		http -
		file {}
		default {
		    set r(path) "$r(scheme):$r(path)"
		    set r(scheme) {}
		}
	    }
	}
    }

    return 1
}

proc DumpURL {varname} {
    upvar $varname r

    puts "r(scheme)=$r(scheme)"
    puts "r(authority)=$r(authority)"
    puts "r(path)=$r(path)"
    puts "r(query)=$r(query)"
    puts "r(fragment)=$r(fragment)"
}

proc BreakUp {str} {
    set r {}
    set l [string length $str]
    for {set i 0} {$i < $l} {incr i} {
	set c [string index $str $i]
	append r $c
	if {$c=="\}"} {
	    append r "\n"
	}
    }
    return $r
}

proc InPath {which} {
    global env
    global tcl_platform

    switch $tcl_platform(platform) {
	unix {
	    set target ${which}
	    set paths [split $env(PATH) :]
	}
	windows {
	    set target ${which}.exe
	    set paths [split $env(PATH) \;]
	}
    }

    foreach p $paths {
	if {[file executable [file join $p $target]]} {
	    return 1
	}
    }
    return 0
}

proc FTPLog {s msg state} {
    global debug

    if {$debug(tcl,ftp)} {
	DisplayLog "$s $msg $state\n"
    }
}

proc HTTPLog {token} {
    global http
    global debug

    if {$debug(tcl,http)} {
	upvar #0 $token t

	DisplayLog "url: $t(url)\n"
	DisplayLog "http: $t(http)\n"
	DisplayLog "type: $t(type)\n"
	DisplayLog "currentsize: $t(currentsize)\n"
	DisplayLog "totalsize: $t(totalsize)\n"
	DisplayLog "status: $t(status)\n"
	if [info exists t(error)] {
	    DisplayLog "error: $t(error)\n"
	}
	DisplayLog "meta: [BreakUp $t(meta)]\n"
    }
}

proc ConfigHTTP {} {
    global http

    # set the User-Agent
    http::config -useragent ds9

    # set the proxy if requested
    if {$http(proxy)} {
	http::config -proxyhost $http(proxy,host) -proxyport $http(proxy,port)
    }
}

proc ProxyHTTP {} {
    global http

    set auth {}
    if {$http(proxy) && $http(auth)} {
	set auth [list "Proxy-Authorization" [concat "Basic" [base64::encode $http(auth,user):$http(auth,passwd)]]]
    } 

    return $auth
}

proc TmpDirOpen {varname} {
    upvar $varname var
    global $varname

    FileLast tmpdirfbox $var
    set var [file dirname [OpenFileDialog tmpdirfbox]]
}

proc DumpCallStack {} {
    for {set x [expr [info level]-1]} {$x>0} {incr x -1} {
	puts "$x: [info level $x]"
    }
}

proc DumpArray {varname} {
    upvar $varname var
    global $varname
    foreach f [array names $varname] {
	puts "${varname}($f) = $var($f)"
    }
}

proc ValidFits {filename} {
    if {$filename == {}} {
	return 0
    }

    # determine if its a fits file
    # first, strip the filename
    if {![regexp -nocase {(.*)(\[.*\])} $filename foo base ext]} {
	set base $filename
	set ext {}
    }

    set fd [open $base]
    if {$fd != {}} {
	set l [read $fd 9]
	close $fd
    }

    if {$l == "SIMPLE  ="} {
	return 1
    } else {
	return 0
    }
}

# Process Cmds

proc ProcessSourceCmd {varname iname} {
    upvar $varname var
    upvar $iname i

    # we need to be realized
    # you never know what someone will try to do
    RealizeDS9

    set fn [lindex $var $i]
    uplevel #0 "source $fn"
}

proc ProcessIconifyCmd {varname iname} {
    upvar $varname var
    upvar $iname i

    global ds9

    switch -- [string tolower [lindex $var $i]] {
	yes -
	true -
	1 {wm iconify $ds9(top)}

	no -
	false -
	0 {wm deiconify $ds9(top)}

	default {
	    wm iconify $ds9(top)
	    incr i -1
	}
    }
}

proc ProcessLowerCmd {varname iname} {
    upvar $varname var
    upvar $iname i

    global ds9
    lower $ds9(top)
}

proc ProcessRaiseCmd {varname iname} {
    upvar $varname var
    upvar $iname i

    global ds9
    raise $ds9(top)
}
